from abc import ABC
from enum import Enum
from typing import Optional, Union
from chex._src.pytypes import PRNGKey
from gymnax.environments import spaces
from gymnax.environments.spaces import Space
from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.environment.utils import (
    convert_continuous_actions,
    convert_discrete_actions,
    convert_multi_discrete_actions,
)


import chex
import jax
import jax.numpy as jnp
import numpy as np

from kinetix.render.renderer_pixels import make_render_pixels_rl
from kinetix.render.renderer_symbolic_entity import make_render_entities
from kinetix.render.renderer_symbolic_flat import make_render_symbolic


class ActionType(Enum):
    CONTINUOUS = 0
    DISCRETE = 1
    MULTI_DISCRETE = 2

    @staticmethod
    def from_string(s: str):
        return {
            "continuous": ActionType.CONTINUOUS,
            "discrete": ActionType.DISCRETE,
            "multi_discrete": ActionType.MULTI_DISCRETE,
        }[s]


class ObservationType(Enum):
    PIXELS = 0
    SYMBOLIC_FLAT = 1
    SYMBOLIC_ENTITY = 2
    BLIND = 3
    SYMBOLIC_FLAT_PADDED = 4

    @staticmethod
    def from_string(s: str):
        return {
            "pixels": ObservationType.PIXELS,
            "symbolic_flat": ObservationType.SYMBOLIC_FLAT,
            "symbolic_entity": ObservationType.SYMBOLIC_ENTITY,
            "blind": ObservationType.BLIND,
            "symbolic_flat_padded": ObservationType.SYMBOLIC_FLAT_PADDED,
        }[s]


class MultiDiscrete(Space):
    def __init__(self, n, number_of_dims_per_distribution):
        self.number_of_dims_per_distribution = number_of_dims_per_distribution
        self.n = n
        self.shape = (number_of_dims_per_distribution.shape[0],)
        self.dtype = jnp.int_

    def sample(self, rng: chex.PRNGKey) -> chex.Array:
        """Sample random action uniformly from set of categorical choices."""
        uniform_sample = jax.random.uniform(rng, shape=self.shape) * self.number_of_dims_per_distribution
        md_dist = jnp.floor(uniform_sample)
        return md_dist.astype(self.dtype)

    def contains(self, x) -> jnp.ndarray:
        """Check whether specific object is within space."""
        range_cond = jnp.logical_and(x >= 0, (x < self.number_of_dims_per_distribution).all())
        return range_cond


class KinetixAction:
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        # This is the processed, unified action space size that is shared with all action types
        # 1 dim per motor and thruster
        self.unified_action_space_size = static_env_params.num_motor_bindings + static_env_params.num_thruster_bindings

    def action_space(self, env_params: Optional[EnvParams] = None) -> Union[spaces.Discrete, spaces.Box]:
        raise NotImplementedError()

    def process_action(self, action: jnp.ndarray, state: EnvState, static_env_params: StaticEnvParams) -> jnp.ndarray:
        raise NotImplementedError()

    def noop_action(self) -> jnp.ndarray:
        raise NotImplementedError()

    def random_action(self, rng: chex.PRNGKey):
        raise NotImplementedError()


class MultiDiscreteActions(KinetixAction):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)

        self.env_params = env_params
        self.static_env_params = static_env_params
        # This is the action space that will be used internally by an agent
        # 3 dims per motor (foward, backward, off) and 2 per thruster (on, off)
        self.n_hot_action_space_size = (
            self.static_env_params.num_motor_bindings * 3 + self.static_env_params.num_thruster_bindings * 2
        )

        def _make_sample_random():
            minval = jnp.zeros(self.unified_action_space_size, dtype=jnp.int32)
            maxval = jnp.ones(self.unified_action_space_size, dtype=jnp.int32) * 3
            maxval = maxval.at[self.static_env_params.num_motor_bindings :].set(2)

            def random(rng):
                return jax.random.randint(rng, shape=(self.unified_action_space_size,), minval=minval, maxval=maxval)

            return random

        self._random = _make_sample_random

        self.number_of_dims_per_distribution = jnp.concatenate(
            [
                np.ones(self.static_env_params.num_motor_bindings) * 3,
                np.ones(self.static_env_params.num_thruster_bindings) * 2,
            ]
        ).astype(np.int32)

    def action_space(self, env_params: Optional[EnvParams] = None) -> MultiDiscrete:
        return MultiDiscrete(self.n_hot_action_space_size, self.number_of_dims_per_distribution)

    def process_action(self, action: jnp.ndarray, state: EnvState, static_env_params: StaticEnvParams) -> jnp.ndarray:
        return convert_multi_discrete_actions(action, state, static_env_params, self.env_params)

    def noop_action(self):
        return jnp.zeros(self.unified_action_space_size, dtype=jnp.int32)

    def random_action(self, rng: chex.PRNGKey):
        return self._random()(rng)


class DiscreteActions(KinetixAction):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)

        self.env_params = env_params
        self.static_env_params = static_env_params

        self._n_actions = (
            self.static_env_params.num_motor_bindings * 2 + 1 + self.static_env_params.num_thruster_bindings
        )

    def action_space(self, env_params: Optional[EnvParams] = None) -> spaces.Discrete:
        return spaces.Discrete(self._n_actions)

    def process_action(self, action: jnp.ndarray, state: EnvState, static_env_params: StaticEnvParams) -> jnp.ndarray:
        return convert_discrete_actions(action, state, static_env_params, self.env_params)

    def noop_action(self) -> int:
        return self.static_env_params.num_motor_bindings * 2

    def random_action(self, rng: chex.PRNGKey):
        return jax.random.randint(rng, shape=(), minval=0, maxval=self._n_actions)


class ContinuousActions(KinetixAction):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)

        self.env_params = env_params
        self.static_env_params = static_env_params

    def action_space(self, env_params: EnvParams | None = None) -> spaces.Discrete | spaces.Box:
        return spaces.Box(
            low=jnp.ones(self.unified_action_space_size) * -1.0,
            high=jnp.ones(self.unified_action_space_size) * 1.0,
            shape=(self.unified_action_space_size,),
        )

    def process_action(self, action: PRNGKey, state: EnvState, static_env_params: StaticEnvParams) -> PRNGKey:
        return convert_continuous_actions(action, state, static_env_params, self.env_params)

    def noop_action(self) -> jnp.ndarray:
        return jnp.zeros(self.unified_action_space_size, dtype=jnp.float32)

    def random_action(self, rng: chex.PRNGKey) -> jnp.ndarray:
        actions = jax.random.uniform(rng, shape=(self.unified_action_space_size,), minval=-1.0, maxval=1.0)
        # Motors between -1 and 1, thrusters between 0 and 1
        actions = actions.at[self.static_env_params.num_motor_bindings :].set(
            jnp.abs(actions[self.static_env_params.num_motor_bindings :])
        )

        return actions


class KinetixObservation(ABC):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        self.env_params = env_params
        self.static_env_params = static_env_params

    def get_obs(self, state: EnvState):
        raise NotImplementedError()

    def observation_space(self, env_params: EnvParams):
        raise NotImplementedError()


class SymbolicPaddedObservations(KinetixObservation):
    def __init__(
        self,
        env_params: EnvParams,
        static_env_params: StaticEnvParams,
    ):
        super().__init__(env_params, static_env_params)
        self.render_function = make_render_symbolic(env_params, static_env_params, True)

    def get_obs(self, state: EnvState):
        return self.render_function(state)


class BlindObservations(KinetixObservation):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)

    def get_obs(self, state: EnvState):
        return jax.nn.one_hot(state.timestep, self.env_params.max_timesteps + 1)


class EntityObservations(KinetixObservation):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams, ignore_mask: bool = False):
        super().__init__(env_params, static_env_params)
        self.render_function = make_render_entities(env_params, static_env_params, ignore_attention_mask=ignore_mask)

    def get_obs(self, state: EnvState):
        return self.render_function(state)

    def observation_space(self, env_params: EnvParams) -> spaces.Dict:
        n_shapes = self.static_env_params.num_polygons + self.static_env_params.num_circles

        def _box(*shape, dtype=jnp.float32, low=-np.inf, high=np.inf):

            return spaces.Box(
                low,
                high,
                shape,
                dtype=dtype,
            )

        return spaces.Dict(
            dict(
                circles=_box(self.static_env_params.num_circles, 19),
                polygons=_box(self.static_env_params.num_polygons, 27),
                joints=_box(self.static_env_params.num_joints * 2, 22),
                thrusters=_box(self.static_env_params.num_thrusters, 8),
                circle_mask=_box(self.static_env_params.num_circles, dtype=bool, low=0, high=1),
                polygon_mask=_box(self.static_env_params.num_polygons, dtype=bool, low=0, high=1),
                joint_mask=_box(self.static_env_params.num_joints * 2, dtype=bool, low=0, high=1),
                thruster_mask=_box(self.static_env_params.num_thrusters, dtype=bool, low=0, high=1),
                attention_mask=_box(4, n_shapes, n_shapes, dtype=bool, low=0, high=1),
                joint_indexes=_box(self.static_env_params.num_joints * 2, 2, dtype=jnp.int32, low=0, high=n_shapes - 1),
                thruster_indexes=_box(self.static_env_params.num_thrusters, dtype=jnp.int32, low=0, high=n_shapes - 1),
            )
        )


class SymbolicObservations(KinetixObservation):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)
        self.render_function = make_render_symbolic(env_params, static_env_params)

    def get_obs(self, state: EnvState):
        return self.render_function(state)

    def observation_space(self, env_params: EnvParams) -> spaces.Box:
        n_shapes = self.static_env_params.num_polygons + self.static_env_params.num_circles
        n_features = (
            (self.static_env_params.num_polygons - 3) * 26
            + self.static_env_params.num_circles * 18
            + self.static_env_params.num_joints * (22 + n_shapes * 2)
            + self.static_env_params.num_thrusters * (8 + n_shapes)
            + 1
        )
        return spaces.Box(
            -np.inf,
            np.inf,
            (n_features,),
            dtype=jnp.float32,
        )


class PixelObservations(KinetixObservation):
    def __init__(self, env_params: EnvParams, static_env_params: StaticEnvParams):
        super().__init__(env_params, static_env_params)
        self.render_function = make_render_pixels_rl(env_params, static_env_params)

    def get_obs(self, state: EnvState):
        return self.render_function(state)

    def observation_space(self, env_params: EnvParams) -> spaces.Box:
        return spaces.Box(
            0.0,
            1.0,
            tuple(a // self.static_env_params.downscale for a in self.static_env_params.screen_dim) + (3,),
            dtype=jnp.float32,
        )



import os
import pathlib

import imageio.v3 as iio
import jax.numpy as jnp
import numpy as np
from PIL import Image


def load_texture(filename, render_size):
    filename = os.path.join(pathlib.Path(__file__).parent.parent.resolve(), "assets", filename)
    img = iio.imread(filename)
    jnp_img = jnp.array(img).astype(jnp.int32)

    if jnp_img.shape[2] == 4:
        jnp_img = jnp_img.at[:, :, 3].set(jnp_img[:, :, 3] // 255)

    img = np.array(jnp_img, dtype=np.uint8)
    image = Image.fromarray(img)
    image = image.resize((render_size, render_size), resample=Image.NEAREST)
    jnp_img = jnp.array(image, dtype=jnp.float32)

    return jnp_img.transpose((1, 0, 2))


EDIT_TEXTURE_RGBA = load_texture("edit.png", 64)
PLAY_TEXTURE_RGBA = load_texture("play.png", 64)

CIRCLE_TEXTURE_RGBA = load_texture("circle.png", 32)
RECT_TEXTURE_RGBA = load_texture("square.png", 32)
TRIANGLE_TEXTURE_RGBA = load_texture("triangle.png", 32)
RJOINT_TEXTURE_6_RGBA = load_texture("rjoint.png", 6)
RJOINT_TEXTURE_RGBA = load_texture("rjoint2.png", 32)

FJOINT_TEXTURE_6_RGBA = load_texture("fjoint.png", 6)
FJOINT_TEXTURE_RGBA = load_texture("fjoint2.png", 32)


ROTATION_TEXTURE_RGBA = load_texture("rotate.png", 32)
SELECT_TEXTURE_RGBA = load_texture("hand.png", 32)

THRUSTER_TEXTURE_RGBA = jnp.rot90(load_texture("thruster6.png", 32), k=3)
THRUSTER_TEXTURE_16_RGBA = jnp.rot90(load_texture("thruster.png", 16), k=3)



from kinetix.render.renderer_pixels import make_render_pixels
from kinetix.render.renderer_symbolic_entity import make_render_entities
from kinetix.render.renderer_symbolic_flat import make_render_symbolic



from kinetix.util.config import (
    generate_params_from_config,
    get_eval_level_groups,
    init_wandb,
    normalise_config,
    get_video_frequency,
)
from kinetix.util.learning import general_eval, no_op_and_random_rollout, sample_trajectories_and_learn
from kinetix.util.saving import (
    load_train_state_from_wandb_artifact_path,
    save_model,
    load_from_json_file,
    export_env_state_to_json,
    get_env_state_from_json,
    save_pickle,
    load_evaluation_levels,
)
from kinetix.util.timing import time_function



from kinetix.environment.env import make_kinetix_env

from kinetix.environment.wrappers import LogWrapper
from kinetix.environment.env_state import EnvState, EnvParams, StaticEnvParams
from kinetix.environment.ued.ued_state import UEDParams

from kinetix.environment.ued.distributions import create_random_starting_distribution

from kinetix.environment.ued.ued import (
    make_mutate_env,
    make_reset_fn_from_config,
    make_vmapped_filtered_level_sampler,
)

from kinetix.environment.ued.distributions import sample_kinetix_level
from kinetix.environment.utils import permute_state, create_empty_env


from kinetix.environment.spaces import (
    ActionType,
    ObservationType,
    PixelObservations,
    SymbolicObservations,
    SymbolicPaddedObservations,
    KinetixObservation,
    KinetixAction,
    ContinuousActions,
    DiscreteActions,
    MultiDiscreteActions,
)



import math
from functools import partial

import jax
import jax.numpy as jnp
from jax2d.engine import PhysicsEngine, calculate_collision_matrix, recalculate_mass_and_inertia, select_shape
from jax2d.sim_state import RigidBody, Thruster

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams


def sample_dimensions(rng, static_env_params: StaticEnvParams, is_rect: bool, ued_params, max_shape_size=None):
    if max_shape_size is None:
        max_shape_size = static_env_params.max_shape_size
    # Returns (half_dimensions, radius)

    rng, _rng = jax.random.split(rng)
    # Don't want overly small shapes
    min_rect_size = 0.05
    min_circle_size = 0.1
    cap_rect = max_shape_size / 2.0 / jnp.sqrt(2.0)
    cap_circ = max_shape_size / 2.0 * ued_params.circle_max_size_coeff
    half_dimensions = (
        jax.lax.select(is_rect, jax.random.uniform(_rng, shape=(2,)), jnp.zeros(2, dtype=jnp.float32))
        * (cap_rect - min_rect_size)
        + min_rect_size
    )

    rng, _rng, __rng = jax.random.split(rng, 3)
    dim_scale = (
        jnp.ones(2)
        .at[jax.random.randint(_rng, shape=(), minval=0, maxval=2)]
        .set(
            jax.lax.select(
                jax.random.uniform(__rng) < ued_params.large_rect_dim_chance, ued_params.large_rect_dim_scale, 1.0
            )
        )
    )
    half_dimensions *= dim_scale

    vertices = jnp.array(
        [
            half_dimensions * jnp.array([1, 1]),
            half_dimensions * jnp.array([1, -1]),
            half_dimensions * jnp.array([-1, -1]),
            half_dimensions * jnp.array([-1, 1]),
        ]
    )

    rng, _rng = jax.random.split(rng)
    radius = (
        jax.lax.select(is_rect, jnp.zeros((), dtype=jnp.float32), jax.random.uniform(_rng, shape=()))
        * (cap_circ - min_circle_size)
        + min_circle_size
    )
    return vertices, half_dimensions, radius


def count_roles(state: EnvState, static_env_params: StaticEnvParams, role: int, include_static_polys=True) -> int:
    active_to_use = state.polygon.active
    if not include_static_polys:
        active_to_use = active_to_use.at[: static_env_params.num_static_fixated_polys].set(False)
    return ((state.polygon_shape_roles == role) * active_to_use).sum() + (
        (state.circle_shape_roles == role) * state.circle.active
    ).sum()


def random_position_on_triangle(rng, vertices):
    verts = vertices[:3]
    rng, _rng, _rng2 = jax.random.split(rng, 3)
    f1 = jax.random.uniform(_rng)
    f2 = jax.random.uniform(_rng2)
    # https://www.reddit.com/r/godot/comments/mqp29g/how_do_i_get_a_random_position_inside_a_collision/
    return verts[0] + jnp.sqrt(f1) * (-verts[0] + verts[1] + f2 * (verts[2] - verts[1]))


def random_position_on_rectangle(rng, vertices):
    verts = vertices[:4]
    rng, _rng, _rng2 = jax.random.split(rng, 3)
    f1 = jax.random.uniform(_rng)
    f2 = jax.random.uniform(_rng2)

    min_x, max_x = jnp.min(verts[:, 0]), jnp.max(verts[:, 0])
    min_y, max_y = jnp.min(verts[:, 1]), jnp.max(verts[:, 1])
    random_x_pos = min_x + f1 * (max_x - min_x)
    random_y_pos = min_y + f2 * (max_y - min_y)

    return jnp.array([random_x_pos, random_y_pos])


def random_position_on_polygon(rng, vertices, n_vertices, static_env_params: StaticEnvParams):
    assert static_env_params.max_polygon_vertices <= 4, "Only supports up to 4 vertices"
    return jax.lax.select(
        n_vertices <= 3, random_position_on_triangle(rng, vertices), random_position_on_rectangle(rng, vertices)
    )


def random_position_on_circle(rng, radius, on_centre_chance):
    rngs = jax.random.split(rng, 3)

    on_centre = jax.random.uniform(rngs[0]) < on_centre_chance

    local_joint_position_circle_theta = jax.random.uniform(rngs[1], shape=()) * 2 * math.pi
    local_joint_position_circle_r = jax.random.uniform(rngs[2], shape=()) * radius
    local_joint_position_circle = jnp.array(
        [
            local_joint_position_circle_r * jnp.cos(local_joint_position_circle_theta),
            local_joint_position_circle_r * jnp.sin(local_joint_position_circle_theta),
        ]
    )

    return jax.lax.select(on_centre, jnp.array([0.0, 0.0]), local_joint_position_circle)


def get_role(rng, state: EnvState, static_env_params: StaticEnvParams, initial_p=None) -> int:

    if initial_p is None:
        initial_p = jnp.array([1.0, 1.0, 1.0, 1.0])

    needs_ball = count_roles(state, static_env_params, 1) == 0
    needs_goal = count_roles(state, static_env_params, 2) == 0
    needs_lava = count_roles(state, static_env_params, 3) == 0

    # always put goal/ball first.
    prob_of_something_else = (needs_ball == 0) & (needs_goal == 0)
    p = initial_p * jnp.array(
        [prob_of_something_else, needs_ball, needs_goal, prob_of_something_else * needs_lava / 3]
    )  # This ensures we cannot more than one ball or goal.
    return jax.random.choice(rng, jnp.array([0, 1, 2, 3]), p=p)


def is_space_for_shape(state: EnvState):
    return jnp.logical_not(jnp.concatenate([state.polygon.active, state.circle.active])).sum() > 0


def is_space_for_joint(state: EnvState):
    return jnp.logical_not(state.joint.active).sum() > 0


def are_there_shapes_present(state: EnvState, static_env_params: StaticEnvParams):
    m = (
        jnp.concatenate([state.polygon.active, state.circle.active])
        .at[: static_env_params.num_static_fixated_polys]
        .set(False)
    )
    return m.sum() > 0


@partial(jax.jit, static_argnums=(2, 9))
def add_rigidbody_to_state(
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    position: jnp.ndarray,
    vertices: jnp.ndarray,
    n_vertices: int,
    radius: float,
    shape_role: int,
    density: float = 1,
    is_circle: bool = False,
):

    new_rigid_body = RigidBody(
        position=position,
        velocity=jnp.array([0.0, 0.0]),
        inverse_mass=1.0,
        inverse_inertia=1.0,
        rotation=0.0,
        angular_velocity=0.0,
        radius=radius,
        active=True,
        friction=1.0,
        vertices=vertices,
        n_vertices=n_vertices,
        collision_mode=1,
        restitution=0.0,
    )

    if is_circle:
        actives = state.circle.active
    else:
        actives = state.polygon.active

    idx = jnp.argmin(actives)

    def noop(state):
        return state

    def replace(state):
        add_func = lambda all, new: all.at[idx].set(new)
        if is_circle:
            state = state.replace(
                circle=jax.tree.map(add_func, state.circle, new_rigid_body),
                circle_densities=state.circle_densities.at[idx].set(density),
                circle_shape_roles=state.circle_shape_roles.at[idx].set(shape_role),
            )
        else:
            state = state.replace(
                polygon=jax.tree.map(add_func, state.polygon, new_rigid_body),
                polygon_densities=state.polygon_densities.at[idx].set(density),
                polygon_shape_roles=state.polygon_shape_roles.at[idx].set(shape_role),
            )

        state = state.replace(
            collision_matrix=calculate_collision_matrix(static_env_params, state.joint),
        )

        state = recalculate_mass_and_inertia(state, static_env_params, state.polygon_densities, state.circle_densities)
        return state

    return jax.lax.cond(jnp.logical_not(actives).sum() > 0, replace, noop, state)


def rectangle_vertices(half_dim):
    return jnp.array(
        [
            half_dim * jnp.array([1, 1]),
            half_dim * jnp.array([1, -1]),
            half_dim * jnp.array([-1, -1]),
            half_dim * jnp.array([-1, 1]),
        ]
    )


# More Manual Control
@partial(jax.jit, static_argnums=(2,))
def add_rectangle_to_state(
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    position: jnp.ndarray,
    width: float,
    height: float,
    shape_role: int,
    density: float = 1,
):

    return add_rigidbody_to_state(
        state,
        env_params,
        static_env_params,
        position,
        rectangle_vertices(jnp.array([width, height]) / 2),
        4,
        0.0,
        shape_role,
        density,
        is_circle=False,
    )


@partial(jax.jit, static_argnums=(2,))
def add_circle_to_state(
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    position: jnp.ndarray,
    radius: float,
    shape_role: int,
    density: float = 1,
):
    return add_rigidbody_to_state(
        state,
        env_params,
        static_env_params,
        position,
        jnp.array([0.0, 0.0]),
        0,
        radius,
        shape_role,
        density,
        is_circle=True,
    )


@partial(jax.jit, static_argnums=(2,))
def add_thruster_to_object(
    state: EnvState,
    env_params: EnvParams,
    static_env_params: StaticEnvParams,
    shape_index: int,
    rotation: float,
    colour: int,
    thruster_power_multiplier: float,
):
    def dummy(state):
        return state

    def do_add(state: EnvState):
        thruster_idx = jnp.argmin(state.thruster.active)

        shape = select_shape(state, shape_index, static_env_params)

        thruster = Thruster(
            object_index=shape_index,
            active=True,
            relative_position=jnp.array([0.0, 0.0]),  # a bit of a hack but reasonable.
            rotation=rotation,
            power=1.0 / jax.lax.select(shape.inverse_mass == 0, 1.0, shape.inverse_mass) * thruster_power_multiplier,
            global_position=select_shape(state, shape_index, static_env_params).position,
        )

        state = state.replace(
            thruster=jax.tree.map(lambda y, x: y.at[thruster_idx].set(x), state.thruster, thruster),
            thruster_bindings=state.thruster_bindings.at[thruster_idx].set(colour),
        )

        return state

    return jax.lax.cond(
        (select_shape(state, shape_index, static_env_params).active)
        & (jnp.logical_not(state.thruster.active).sum() > 0),
        do_add,
        dummy,
        state,
    )


def make_velocities_zero(state: EnvState):
    def inner(state):
        return state.replace(
            polygon=state.polygon.replace(
                angular_velocity=state.polygon.angular_velocity * 0,
                velocity=state.polygon.velocity * 0,
            ),
            circle=state.circle.replace(
                angular_velocity=state.circle.angular_velocity * 0,
                velocity=state.circle.velocity * 0,
            ),
        )

    return inner(state)


def make_do_dummy_step(
    env_params: EnvParams, static_sim_params: StaticEnvParams, zero_collisions=True, zero_velocities=True
):
    env = PhysicsEngine(static_sim_params)

    @jax.jit
    def _step_fn(state):
        state, _ = env.step(
            state, env_params, jnp.zeros((static_sim_params.num_joints + static_sim_params.num_thrusters,))
        )
        return state

    def do_dummy_step(state: EnvState) -> EnvState:
        rng = jax.random.PRNGKey(0)
        og_col = state.collision_matrix
        g = state.gravity
        state = state.replace(
            collision_matrix=state.collision_matrix & (not zero_collisions), gravity=state.gravity * 0
        )
        state = _step_fn(state)
        state = state.replace(gravity=g, collision_matrix=og_col)
        if zero_velocities:
            state = make_velocities_zero(state)
        return state

    return do_dummy_step



from typing import Callable
import jax
import jax.numpy as jnp
import numpy as np

from kinetix.environment.env_state import EnvParams, EnvState, StaticEnvParams
from kinetix.render.renderer_symbolic_common import (
    make_circle_features,
    make_joint_features,
    make_polygon_features,
    make_thruster_features,
)
from kinetix.environment.utils import create_empty_env
from kinetix.render.inverse_renderer_common import features_to_joint, features_to_rigidbody, features_to_thruster
from jax2d.engine import calculate_collision_matrix


def _get_symbolic_features(state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams):
    nshapes = static_env_params.num_polygons + static_env_params.num_circles

    polygon_features, polygon_mask = make_polygon_features(state, env_params, static_env_params)
    mask_to_ignore_walls_ceiling = np.ones(static_env_params.num_polygons, dtype=bool)
    mask_to_ignore_walls_ceiling[np.array([i for i in range(1, static_env_params.num_static_fixated_polys)])] = False

    polygon_features = polygon_features[mask_to_ignore_walls_ceiling]
    polygon_mask = polygon_mask[mask_to_ignore_walls_ceiling]

    circle_features, circle_mask = make_circle_features(state, env_params, static_env_params)
    joint_features, joint_idxs, joint_mask = make_joint_features(state, env_params, static_env_params)
    thruster_features, thruster_idxs, thruster_mask = make_thruster_features(state, env_params, static_env_params)

    two_J = joint_features.shape[0]
    J = two_J // 2  # for symbolic only have the one
    joint_features = jnp.concatenate(
        [
            joint_features[:J],  # shape (2 * J, K)
            (jax.nn.one_hot(joint_idxs[:J, 0], nshapes)),  # shape (2 * J, N)
            (jax.nn.one_hot(joint_idxs[:J, 1], nshapes)),  # shape (2 * J, N)
        ],
        axis=1,
    )
    thruster_features = jnp.concatenate(
        [
            thruster_features,
            (jax.nn.one_hot(thruster_idxs, nshapes)),
        ],
        axis=1,
    )

    polygon_features = jnp.where(polygon_mask[:, None], polygon_features, 0.0)
    circle_features = jnp.where(circle_mask[:, None], circle_features, 0.0)
    joint_features = jnp.where(joint_mask[:J, None], joint_features, 0.0)
    thruster_features = jnp.where(thruster_mask[:, None], thruster_features, 0.0)

    return nshapes, polygon_features, circle_features, joint_features, thruster_features, joint_idxs, thruster_idxs


def make_render_symbolic(
    env_params: EnvParams, static_env_params: StaticEnvParams, padded: bool = False, clip: bool = True
) -> Callable[[EnvState], jnp.ndarray]:
    def render_symbolic(state: EnvState) -> jnp.ndarray:
        _, polygon_features, circle_features, joint_features, thruster_features, _, _ = _get_symbolic_features(
            state, env_params, static_env_params
        )
        if padded:
            # pad final dimension with zeros to make all length max_width
            max_width = max(
                polygon_features.shape[-1],
                circle_features.shape[-1],
                joint_features.shape[-1],
                thruster_features.shape[-1],
            )
            polygon_features = jnp.pad(
                polygon_features, ((0, 0), (0, max_width - polygon_features.shape[-1]))
            )  # (2, 36)
            circle_features = jnp.pad(circle_features, ((0, 0), (0, max_width - circle_features.shape[-1])))  # (2, 36)
            joint_features = jnp.pad(joint_features, ((0, 0), (0, max_width - joint_features.shape[-1])))  # (1, 36)
            thruster_features = jnp.pad(
                thruster_features, ((0, 0), (0, max_width - thruster_features.shape[-1]))
            )  # (1, 36)
            # stack
            obs = jnp.concatenate(
                [polygon_features, circle_features, joint_features, thruster_features], axis=0
            )  # (6, 36)
            # add one-hot encoding of the shape type (+4 to feature dim)
            n_polys, n_circles, n_joints, n_thrusters = (
                polygon_features.shape[0],
                circle_features.shape[0],
                joint_features.shape[0],
                thruster_features.shape[0],
            )
            object_types = jnp.array([0] * n_polys + [1] * n_circles + [2] * n_joints + [3] * n_thrusters)  # (6,)
            one_hot_object_types = jax.nn.one_hot(object_types, 4)  # (6, 4)
            obs = jnp.concatenate([obs, one_hot_object_types], axis=-1)  # (6, 40)
            # add gravity (+1 to feature dim)
            gravity = jnp.full((obs.shape[0], 1), state.gravity[1] / 10)  # (6, 1)
            obs = jnp.concatenate([obs, gravity], axis=-1)  # (6, 41)
            # clip
            if clip:
                obs = jnp.clip(obs, min=-10.0, max=10.0)
            obs = jnp.nan_to_num(obs)
            return obs

        else:
            obs = jnp.concatenate(
                [
                    polygon_features.flatten(),
                    circle_features.flatten(),
                    joint_features.flatten(),
                    thruster_features.flatten(),
                    jnp.array([state.gravity[1]]) / 10,
                ],
                axis=0,
            )
            if clip:
                obs = jnp.clip(obs, min=-10.0, max=10.0)
            obs = jnp.nan_to_num(obs)
            return obs

    return render_symbolic


def make_inverse_render_symbolic(
    env_state: EnvState, env_params: EnvParams, static_env_params: StaticEnvParams
) -> Callable[[jnp.ndarray], EnvState]:
    """This creates an inverse renderer, which takes in an observation and returns an EnvState.

        Note: Since the symbolic observation clips values to be between -10 and 10, the inverse renderer will not be able to
        perfectly reconstruct the original EnvState.
    Args:
        env_state (EnvState): This is a dummy env state, and is used to get the shapes of the various arrays.

    Returns:
        Callable[[jnp.ndarray], EnvState]: Maps symbolic observation to EnvState.
    """
    (
        nshapes,
        polygon_features_dummy,
        circle_features_dummy,
        joint_features_dummy,
        thruster_features_dummy,
        _,
        _,
    ) = _get_symbolic_features(env_state, env_params, static_env_params)

    def symbolic_flat_obs_to_env_state(obs: jnp.ndarray) -> EnvState:
        # polygons
        start_index = 0
        end_index = polygon_features_dummy.shape[0] * polygon_features_dummy.shape[1]
        polygon_features = jnp.reshape(obs[start_index:end_index], polygon_features_dummy.shape)

        # circles
        start_index = end_index
        end_index += circle_features_dummy.shape[0] * circle_features_dummy.shape[1]
        circle_features = jnp.reshape(obs[start_index:end_index], circle_features_dummy.shape)

        # joints
        start_index = end_index
        end_index += joint_features_dummy.shape[0] * joint_features_dummy.shape[1]
        joint_features = jnp.reshape(obs[start_index:end_index], joint_features_dummy.shape)

        # thrusters
        start_index = end_index
        end_index += thruster_features_dummy.shape[0] * thruster_features_dummy.shape[1]
        thruster_features = jnp.reshape(obs[start_index:end_index], thruster_features_dummy.shape)

        env_state = create_empty_env(static_env_params)

        # Polygon & Circle
        circle, circle_shape_roles, circle_densities = features_to_rigidbody(
            env_params, static_env_params, circle_features, static_env_params.num_circles, is_circle=True
        )
        # polygon

        polygon, polygon_shape_roles, polygon_densities = features_to_rigidbody(
            env_params, static_env_params, polygon_features, static_env_params.num_polygons - 3, is_circle=False
        )

        def _add_walls_ceiling(small_array, dummy_large_array):
            return jnp.concatenate(
                [small_array[:1], dummy_large_array[1 : static_env_params.num_static_fixated_polys], small_array[1:]],
                axis=0,
            )

        (polygon, polygon_shape_roles, polygon_densities) = jax.tree.map(
            _add_walls_ceiling,
            (polygon, polygon_shape_roles, polygon_densities),
            (env_state.polygon, env_state.polygon_shape_roles, env_state.polygon_densities),
        )
        env_state = env_state.replace(
            circle=circle,
            polygon=polygon,
            circle_shape_roles=circle_shape_roles,
            polygon_shape_roles=polygon_shape_roles,
            circle_densities=circle_densities,
            polygon_densities=polygon_densities,
        )

        # Thruster

        thrusters, thruster_bindings = features_to_thruster(static_env_params, env_state, thruster_features, nshapes)

        env_state = env_state.replace(
            thruster=thrusters,
            thruster_bindings=thruster_bindings,
        )

        # Joint
        joint, motor_bindings = features_to_joint(static_env_params, env_state, joint_features, nshapes)
        env_state = env_state.replace(
            joint=joint,
            motor_bindings=motor_bindings,
        )

        return env_state.replace(
            collision_matrix=calculate_collision_matrix(static_env_params, env_state.joint),
        )

    return symbolic_flat_obs_to_env_state



from functools import partial
from typing import Any, Optional, Union

import chex
import jax
from flax import struct

from kinetix.environment.env_state import EnvState


class GymnaxWrapper(object):
    """Base class for Gymnax wrappers."""

    def __init__(self, env):
        self._env = env

    # provide proxy access to regular attributes of wrapped object
    def __getattr__(self, name):
        return getattr(self._env, name)


@struct.dataclass
class LogEnvState:
    env_state: Any
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int
    timestep: int


class LogWrapper(GymnaxWrapper):
    """Log the episode returns and lengths."""

    def __init__(self, env):
        super().__init__(env)

    @partial(jax.jit, static_argnums=(0, 2))
    def reset(self, key: chex.PRNGKey, env_params=None, override_reset_state: Optional[EnvState] = None):
        obs, env_state = self._env.reset(key, env_params, override_reset_state)
        state = LogEnvState(env_state, 0.0, 0, 0.0, 0, 0)
        return obs, state

    @partial(jax.jit, static_argnums=(0, 4))
    def step(
        self,
        key: chex.PRNGKey,
        state,
        action: Union[int, float],
        env_params=None,
        override_reset_state: Optional[EnvState] = None,
    ):
        if isinstance(override_reset_state, LogEnvState):
            override_reset_state = override_reset_state.env_state
        obs, env_state, reward, done, info = self._env.step(
            key, state.env_state, action, env_params, override_reset_state
        )
        new_episode_return = state.episode_returns + reward
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - done),
            episode_lengths=new_episode_length * (1 - done),
            returned_episode_returns=state.returned_episode_returns * (1 - done) + new_episode_return * done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - done) + new_episode_length * done,
            timestep=state.timestep + 1,
        )
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["returned_episode_solved"] = info["GoalR"]
        info["timestep"] = state.timestep
        info["returned_episode"] = done
        return obs, state, reward, done, info

    def __hash__(self):
        return hash(self._env)

    def __eq__(self, value):
        return hash(self) == hash(value)



import copy
import datetime
import gzip
import json
import os
from collections import defaultdict
from hashlib import md5
from typing import List, Tuple

import jax
import jax.numpy as jnp
import numpy as np
from omegaconf import OmegaConf

import wandb
from kinetix.environment.env_state import EnvParams, StaticEnvParams
from kinetix.environment.spaces import ActionType
from kinetix.environment.spaces import ObservationType
from kinetix.environment.ued.ued_state import UEDParams
from kinetix.util.saving import get_correct_path_of_json_level, load_from_json_file


def get_hash_without_seed(config):
    old_seed = config["seed"]
    config["seed"] = 0
    ans = md5(OmegaConf.to_yaml(config, sort_keys=True).encode()).hexdigest()
    config["seed"] = old_seed
    return ans


def get_date() -> str:
    return datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")


def generate_params_from_config(config):
    if config.get("env_size_type", "predefined") == "custom":
        # must load env params from a file
        _, static_env_params, env_params = load_from_json_file(config["custom_path"])
        return env_params, static_env_params.replace(
            frame_skip=config["frame_skip"],
        )
    env_params = EnvParams().replace(dense_reward_scale=config["dense_reward_scale"])

    static_env_params = StaticEnvParams().replace(
        num_polygons=config["num_polygons"],
        num_circles=config["num_circles"],
        num_joints=config["num_joints"],
        num_thrusters=config["num_thrusters"],
        frame_skip=config["frame_skip"],
        num_motor_bindings=config["num_motor_bindings"],
        num_thruster_bindings=config["num_thruster_bindings"],
    )

    return env_params, static_env_params


def get_eval_level_groups(eval_levels: List[str]) -> List[Tuple[str, str]]:
    def get_groups(s):
        # This is the size group
        group_one = s.split("/")[0]
        group_two = s.split("/")[1].split("_")[0]
        group_two = "".join([i for i in group_two if not i.isdigit()])
        if group_two == "h":
            group_two = "handmade"
        if group_two == "r":
            group_two = "random"
        return f"{group_one}_all", f"{group_one}_{group_two}"

    indices = defaultdict(list)

    for idx, s in enumerate(eval_levels):
        groups = get_groups(s)
        for group in groups:
            indices[group].append(idx)

    indices2 = {}
    for g in indices:
        indices2[g] = np.array(indices[g])

    return indices2


def normalise_config(config, name, editor_config=False, save_config: bool = True):
    old_config = copy.deepcopy(config)
    keys = ["env", "learning", "model", "misc", "eval", "ued", "env_size", "train_levels"]
    for k in keys:
        if k not in config:
            config[k] = {}
        small_d = config[k]
        del config[k]
        for kk, vv in small_d.items():
            assert kk not in config, kk
            config[kk] = vv

    config["observation_type_str"] = config["observation_type"]
    config["observation_type"] = ObservationType.from_string(config["observation_type"])
    config["action_type_str"] = config["action_type"]
    config["action_type"] = ActionType.from_string(config["action_type"])
    if not editor_config:
        config[
            "env_name"
        ] = f"Kinetix-{config['observation_type_str'].title().replace('_', '')}-{config['action_type_str'].title().replace('_', '')}"
        if config["num_train_envs"] == 2048 and config["observation_type"] == ObservationType.PIXELS:
            config["num_train_envs"] = 512
        if "SFL" in name and config["env_size_name"] in ["m", "l"]:
            config["eval_num_attempts"] = 6  # to avoid a very weird XLA bug.
        config["hash"] = get_hash_without_seed(config)

        config["random_hash"] = np.random.randint(2**31)

        config["log_save_path"] = f"logs/{config['hash']}/{config['seed']}-{get_date()}"
        if save_config:
            os.makedirs(config["log_save_path"], exist_ok=True)
            with open(f"{config['log_save_path']}/config.yaml", "w") as f:
                f.write(OmegaConf.to_yaml(old_config))
        if config["group"] == "auto":
            config["group"] = f"{name}-" + config["group_auto_prefix"] + config["env_name"].replace("Kinetix-", "")
            config["group"] += "-" + str(config["env_size_name"])

        if config["eval_levels"] == ["auto"] or config["eval_levels"] == "auto":
            config["eval_levels"] = config["train_levels_list"]
            print("Using Auto eval levels:", config["eval_levels"])
        config["num_eval_levels"] = len(config["eval_levels"])

        steps = (
            config["num_steps"]
            * config.get("outer_rollout_steps", 1)
            * config["num_train_envs"]
            * (2 if name == "PAIRED" else 1)
        )
        config["num_updates"] = int(config["total_timesteps"]) // steps

        nsteps = int(config["total_timesteps"] // 1e6)
        letter = "M"
        if nsteps >= 1000:
            nsteps = nsteps // 1000
            letter = "B"
        config["run_name"] = (
            config["env_name"] + f"-{name}-" + str(nsteps) + letter + "-" + str(config["num_train_envs"])
        )

        if config["checkpoint_save_freq"] >= config["num_updates"]:
            config["checkpoint_save_freq"] = config["num_updates"]
    return config


def get_tags(config, name):
    return [name]
    tags = [name]
    if name in ["PLR", "ACCEL", "DR"]:
        if config["use_accel"]:
            tags.append("ACCEL")
        else:
            tags.append("PLR")
    return tags


def init_wandb(config, name) -> wandb.run:
    run = wandb.init(
        config=config,
        project=config["wandb_project"],
        group=config["group"],
        name=config["run_name"],
        entity=config["wandb_entity"],
        mode=config["wandb_mode"],
        tags=get_tags(config, name),
    )
    wandb.define_metric("timing/num_updates")
    wandb.define_metric("timing/num_env_steps")
    wandb.define_metric("*", step_metric="timing/num_env_steps")
    wandb.define_metric("timing/sps", step_metric="timing/num_env_steps")
    return run


def save_data_to_local_file(data_to_save, config):
    if not config.get("save_local_data", False):
        return

    def reverse_in(li, value):
        for i, v in enumerate(li):
            if v in value:
                return True
        return False

    clean_data = {k: v for k, v in data_to_save.items() if not reverse_in(["media/", "images/"], k)}

    def _clean(x):
        if isinstance(x, jnp.ndarray):
            return x.tolist()
        elif isinstance(x, jnp.float32):
            if jnp.isnan(x):
                return -float("inf")
            return round(float(x) * 1000) / 1000
        elif isinstance(x, jnp.int32):
            return int(x)
        return x

    clean_data = jax.tree.map(lambda x: _clean(x), clean_data)
    print("Saving this data:", clean_data)
    with open(f"{config['log_save_path']}/data.jsonl", "a+") as f:
        f.write(json.dumps(clean_data) + "\n")


def compress_log_files_after_run(config):
    fpath = f"{config['log_save_path']}/data.jsonl"
    with open(fpath, "rb") as f_in, gzip.open(fpath + ".gz", "wb") as f_out:
        f_out.writelines(f_in)


def get_video_frequency(config, update_step):
    frac_through_training = update_step / config["num_updates"]
    vid_frequency = (
        config["eval_freq"]
        * config["video_frequency"]
        * jax.lax.select(
            (0.1 <= frac_through_training) & (frac_through_training < 0.3),
            1,
            jax.lax.select(
                (0.3 <= frac_through_training) & (frac_through_training < 0.6),
                2,
                4,
            ),
        )
    )
    return vid_frequency



import os
import sys
import time
import tkinter
import tkinter.filedialog
from enum import Enum
from timeit import default_timer as tmr
from tkinter import Tk

import hydra
import jax
import jax.numpy as jnp
import numpy as np
import optax
import pygame
import pygame_widgets
from flax.serialization import to_state_dict
from flax.training.train_state import TrainState
from jax2d.engine import (
    calc_inverse_inertia_circle,
    calc_inverse_inertia_polygon,
    calc_inverse_mass_circle,
    calc_inverse_mass_polygon,
    calculate_collision_matrix,
    recalculate_mass_and_inertia,
    recompute_global_joint_positions,
    select_shape,
)
from jax2d.maths import rmat
from jax2d.sim_state import RigidBody
from omegaconf import OmegaConf
from PIL import Image
from pygame_widgets.slider import Slider
from pygame_widgets.textbox import TextBox
from pygame_widgets.toggle import Toggle

from kinetix.environment import (
    EnvParams,
    EnvState,
    StaticEnvParams,
    UEDParams,
    create_empty_env,
    make_kinetix_env,
    permute_state,
)
from kinetix.environment import ActionType, ObservationType
from kinetix.environment.ued.mutators import (
    make_mutate_change_shape_rotation,
    make_mutate_change_shape_size,
    mutate_add_connected_shape,
    mutate_add_shape,
    mutate_add_thruster,
    mutate_change_shape_location,
    mutate_remove_joint,
    mutate_remove_shape,
    mutate_remove_thruster,
    mutate_swap_role,
    mutate_toggle_fixture,
)
from kinetix.environment.ued.ued import ALL_MUTATION_FNS, make_mutate_env
from kinetix.environment.ued.util import rectangle_vertices
from kinetix.models import ScannedRNN, make_network_from_config
from kinetix.render import make_render_entities, make_render_pixels
from kinetix.render.textures import (
    CIRCLE_TEXTURE_RGBA,
    EDIT_TEXTURE_RGBA,
    PLAY_TEXTURE_RGBA,
    RECT_TEXTURE_RGBA,
    RJOINT_TEXTURE_RGBA,
    SELECT_TEXTURE_RGBA,
    THRUSTER_TEXTURE_RGBA,
    TRIANGLE_TEXTURE_RGBA,
)
from kinetix.util import (
    export_env_state_to_json,
    generate_params_from_config,
    get_env_state_from_json,
    load_from_json_file,
    load_train_state_from_wandb_artifact_path,
    normalise_config,
    save_pickle,
    time_function,
)

jax.config.update("jax_compilation_cache_dir", ".cache-location")


# sys.path.append("editor")


from sys import platform

# Hack for macOS
if platform == "darwin":
    root = Tk()
    root.destroy()

editor = None
outer_timer = tmr()
EMPTY_ENV = False


@jax.jit
def _signed_line_distance(a, b, c):
    return (b[0] - a[0]) * (c[1] - a[1]) - (b[1] - a[1]) * (c[0] - a[0])


@jax.jit
def _mouse_dist_to_vertices(mpos, i1, i2, vertices):
    return _signed_line_distance(mpos, vertices[i1], vertices[i2])


class ObjectType(Enum):
    POLYGON = 0
    CIRCLE = 1
    JOINT = 2
    THRUSTER = 3


class EditMode(Enum):
    ADD_CIRCLE = 0
    ADD_RECTANGLE = 1
    ADD_JOINT = 2
    SELECT = 3
    ADD_TRIANGLE = 4
    ADD_THRUSTER = 5


TOTAL_DUMMY_STEPS_TO_SNAP = 0
SNAPPING_DIST = 0.1


def select_object(state: EnvState, type: int, index: int):
    if type is None:
        type = ObjectType.POLYGON
    li = {0: state.polygon, 1: state.circle, 2: state.joint, 3: state.thruster}[type.value]
    return jax.tree.map(lambda x: x[index], li)


def snap_to_center(shape: RigidBody, position: jnp.ndarray):
    if jnp.linalg.norm(shape.position - position) < SNAPPING_DIST:
        return shape.position
    return position


def snap_to_polygon_center_line(polygon: RigidBody, position: jnp.ndarray):
    # Snap to the center line
    r = rmat(polygon.rotation)
    x = jnp.matmul(r, position - polygon.position)
    if jnp.abs(x[0]) < SNAPPING_DIST:
        x = x.at[0].set(0.0)
    if jnp.abs(x[1]) < SNAPPING_DIST:
        x = x.at[1].set(0.0)
    x = jnp.matmul(r.transpose(), x)
    return x + polygon.position


def snap_to_circle_center_line(circle: RigidBody, position: jnp.ndarray):
    # Snap to the center line, i.e. on the edge of the circle, if the position is close enough to directly below the circle, etc., snap the position to that
    x = position - circle.position
    if jnp.linalg.norm(x) < SNAPPING_DIST:
        return circle.position
    angle = (jnp.arctan2(x[1], x[0]) + 2 * jnp.pi) % (2 * jnp.pi)

    for i in range(0, 8):
        if jnp.abs(angle - i * jnp.pi / 4) < jnp.radians(25):  # 25 degrees
            angle = i * jnp.pi / 4
            break
    x = jnp.array([jnp.cos(angle), jnp.sin(angle)]) * circle.radius
    return x + circle.position


def prompt_file(save=False):
    dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "kinetix/levels")
    """Create a Tk file dialog and cleanup when finished"""
    top = tkinter.Tk()
    top.withdraw()  # hide window
    if save:
        file_name = tkinter.filedialog.asksaveasfilename(parent=top, initialdir=dir)
    else:
        file_name = tkinter.filedialog.askopenfilename(parent=top, initialdir=dir)
    top.destroy()
    return file_name


def get_numeric_key_pressed(pygame_events, is_mod=False):
    for event in pygame_events:
        if not is_mod:
            if event.type == pygame.KEYDOWN and event.unicode.isdigit():
                return int(event.unicode)
        else:
            if event.type == pygame.KEYDOWN:
                pass
            if event.type == pygame.KEYDOWN and event.key in [pygame.K_0 + i for i in range(10)]:
                return int(event.key - pygame.K_0)
    return None


def new_env(static_env_params):
    return create_empty_env(static_env_params)


myrng = jax.random.PRNGKey(0)


def make_reset_function(static_env_params):
    def reset(rng):
        return create_empty_env(static_env_params)

    return reset


def new_env(static_env_params):
    global myrng
    if EMPTY_ENV:
        env_state = create_empty_env(static_env_params)
    else:
        env_state = get_env_state_from_json("l/h0_angrybirds.json")
    return env_state


class Editor:
    def __init__(self, env, env_params, config, upscale=1):
        self.env = env
        self.upscale = upscale
        self.env_params = env_params
        self.static_env_params = env.static_env_params
        self.ued_params = UEDParams()

        self.side_panel_width = env.static_env_params.screen_dim[0] // 3

        self.rng = jax.random.PRNGKey(0)
        self.config = config
        self.env_state = new_env(self.static_env_params)

        self.rng, _rng = jax.random.split(self.rng)
        self.play_state = self.env_state
        self.last_played_level = None

        self.pygame_events = []

        self.mutate_world = make_mutate_env(env.static_env_params, env_params, self.ued_params)

        self.num_triangle_clicks = 0
        self.triangle_order = jnp.array([0, 1, 2])

        # Init rendering
        pygame.init()
        pygame.key.set_repeat(250, 75)

        self.screen_surface = pygame.display.set_mode(
            tuple(
                (t + extra) * self.upscale
                for t, extra in zip(self.static_env_params.screen_dim, (self.side_panel_width, 0))
            ),
            display=0,
        )
        self.all_widgets = {}
        self._setup_side_panel()
        self.has_done_action = False

        self._setup_rendering(self.static_env_params, env_params)

        self._render_edit_overlay_fn = jax.jit(self._render_edit_overlay)
        self._step_fn = jax.jit(env.step)

        # Agent
        if self.config["agent_taking_actions"]:
            self._entity_renderer = jax.jit(make_render_entities(env_params, self.env.static_env_params))
            self.network = make_network_from_config(env, env_params, config)

            rng = jax.random.PRNGKey(0)
            dones = jnp.zeros((config["num_train_envs"]), dtype=jnp.bool_)
            rng, _rng = jax.random.split(rng)
            init_hstate = ScannedRNN.initialize_carry(config["num_train_envs"])
            obsv = self._entity_renderer(self.play_state)
            obsv = jax.tree.map(lambda x: jnp.repeat(x[None, ...], repeats=config["num_train_envs"], axis=0), obsv)
            init_x = jax.tree.map(lambda x: x[None, ...], (obsv, dones))
            network_params = self.network.init(_rng, init_hstate, init_x)

            tx = optax.chain(
                optax.clip_by_global_norm(config["max_grad_norm"]),
                optax.adam(config["lr"], eps=1e-5),
            )
            self.train_state = TrainState.create(
                apply_fn=self.network.apply,
                params=network_params,
                tx=tx,
            )

            self.train_state = load_train_state_from_wandb_artifact_path(self.train_state, config["agent_wandb_path"])

            self.apply_fn = jax.jit(self.network.apply)

        # JIT Compile

        def _jit_render():
            self._render_fn(self.play_state)
            self._render_fn_edit(self.play_state)

        time_function(self._jit_step, "_jit_step")
        time_function(_jit_render, "_jit_render")

        # self._step_fn(rng, self.play_state, 0, self.env_params)

        # Editing
        self.is_editing = True

        self.edit_shape_mode = EditMode.ADD_CIRCLE
        self.creating_shape = False
        self.create_shape_position = jnp.array([0.0, 0.0])
        self.creating_shape_index = 0
        self.selected_shape_index = -1
        self.selected_shape_type = ObjectType.POLYGON

        self.all_selected_shapes = []

        self.rng = jax.random.PRNGKey(0)
        time_function(self._jit, "self._jit")
        self._put_state_values_into_gui(self.env_state)

        self.mutate_change_shape_size = make_mutate_change_shape_size(self.env_params, self.static_env_params)
        self.mutate_change_shape_rotation = make_mutate_change_shape_rotation(self.env_params, self.static_env_params)

    def _setup_rendering(self, static_env_params: StaticEnvParams, env_params: EnvParams):
        def _make_render(should_do_edit_additions=False):
            def _render(env_state):
                side_panel = self._render_side_panel()
                render_pixels = make_render_pixels(env_params=env_params, static_params=static_env_params)
                pixels = render_pixels(
                    env_state,
                )
                pixels = jnp.concatenate([side_panel, pixels], axis=0)

                pixels = jnp.repeat(pixels, repeats=static_env_params.downscale * self.upscale, axis=0)
                pixels = jnp.repeat(pixels, repeats=static_env_params.downscale * self.upscale, axis=1)

                return pixels[:, ::-1, :]

            return _render

        def _make_screenshot_render():
            def _render(env_state):
                px_upscale = 4

                static_params = static_env_params.replace(screen_dim=(500 * px_upscale, 500 * px_upscale))
                ss_env_params = env_params.replace(
                    pixels_per_unit=100 * px_upscale,
                )

                render_pixels = make_render_pixels(
                    env_params=ss_env_params,
                    static_params=static_params,
                    render_rjoint_sectors=False,
                    pixel_upscale=2 * px_upscale,
                )
                pixels = render_pixels(
                    env_state,
                )

                return pixels[:, ::-1, :]

            return _render

        self._render_fn_edit = jax.jit(_make_render(True))
        self._render_fn = jax.jit(_make_render(False))
        self._render_fn_screenshot = jax.jit(_make_screenshot_render())

    def _jit(self):
        self._get_circles_on_mouse(self.env_state)
        self._get_polygons_on_mouse(self.env_state)
        self._get_revolute_joints_on_mouse(self.env_state)
        self._get_thrusters_on_mouse(self.env_state)
        self.pygame_events = list(pygame.event.get())
        self._handle_events(do_dummy=True)

        state = self.play_state
        for mutation_fn in ALL_MUTATION_FNS:
            mutation_fn(jax.random.PRNGKey(0), state, self.env_params, self.static_env_params, self.ued_params)

    def _jit_step(self, env_state: EnvState = None):
        rng = jax.random.PRNGKey(0)
        state_to_use = env_state if env_state is not None else self.play_state
        self._step_fn = self._step_fn.lower(
            rng,
            self.env.reset(rng, self.env_params, state_to_use)[1],
            jnp.zeros(
                self.env.static_env_params.num_motor_bindings + self.env.static_env_params.num_thruster_bindings,
                dtype=jnp.int32,
            ),
            self.env_params,
        ).compile()

    def update(self, rng):
        # Update pygame events
        self.pygame_events = list(pygame.event.get())

        for event in self.pygame_events:
            if event.type == pygame.KEYDOWN:
                if event.key == pygame.K_SPACE:
                    self.has_done_action = False
                    self.is_editing = not self.is_editing
                    if not self.is_editing:
                        self.env_state = self._discard_shape_being_created(self.env_state)
                        self.env_state = self._reset_select_shape(self.env_state)
                        self.env_state = self.env_state.replace(
                            collision_matrix=calculate_collision_matrix(self.static_env_params, self.env_state.joint),
                        )
                        self.rng, _rng = jax.random.split(self.rng)
                        self.play_state = self.env_state
                        self.last_played_level = self.play_state
                elif event.key == pygame.K_s and not self.is_editing:
                    self.take_screenshot()

        if self.is_editing:
            self.env_state = self.edit()
        else:
            rng, _rng = jax.random.split(rng)
            # action = []
            action = jnp.zeros(
                self.env.static_env_params.num_motor_bindings + self.env.static_env_params.num_thruster_bindings,
                dtype=jnp.int32,
            )

            keys = pygame.key.get_pressed()
            if keys[pygame.K_LEFT]:
                action = action.at[0].set(1)
            if keys[pygame.K_RIGHT]:
                action = action.at[0].set(2)
            if keys[pygame.K_UP]:
                action = action.at[1].set(1)
            if keys[pygame.K_DOWN]:
                action = action.at[1].set(2)

            if keys[pygame.K_1]:
                action = action.at[0 + self.env.static_env_params.num_motor_bindings].set(1)
            if keys[pygame.K_2]:
                action = action.at[1 + self.env.static_env_params.num_motor_bindings].set(1)
            if keys[pygame.K_3]:
                action = action.at[2 + self.env.static_env_params.num_motor_bindings].set(1)
            if keys[pygame.K_4]:
                action = action.at[3 + self.env.static_env_params.num_motor_bindings].set(1)
            # if self.has_done_action: action = action * 0
            self.has_done_action = self.has_done_action | (action != 0).any()

            if self.config["agent_taking_actions"]:
                obs = self._entity_renderer(self.play_state)
                obs = jax.tree.map(lambda x: x[None, ...], obs)

                last_done = jnp.zeros((1, 1), dtype=bool)
                ac_in = (jax.tree.map(lambda x: x[np.newaxis, :], obs), last_done[np.newaxis, :])
                hstate = ScannedRNN.initialize_carry(1)
                hstate, pi, value = self.apply_fn(self.train_state.params, hstate, ac_in)
                rng, _rng = jax.random.split(rng)
                action = pi.sample(seed=_rng)
                action = action[0, 0]

            _rng, __rng = jax.random.split(_rng)
            obs, self.play_state, reward, done, info = self._step_fn(
                _rng, self.env.reset(__rng, self.env_params, self.play_state)[1], action, self.env_params
            )
            if done:
                self.rng, _rng = jax.random.split(self.rng)
                self.play_state = self.env_state
        state_to_render = self.env_state if self.is_editing else self.play_state

        self.render(state_to_render)
        self._handle_events()

        # Update screen
        pygame.display.flip()

    def take_screenshot(self):
        print("screenshot!")

        pixels = self._render_fn_screenshot(self.play_state)
        mtime = round(time.time() * 1000)

        pixels = pixels.transpose((1, 0, 2))

        # Black border
        border_thickness = 5
        pixels = pixels.at[:, :border_thickness].set(0.0)
        pixels = pixels.at[:, -border_thickness:].set(0.0)
        pixels = pixels.at[:border_thickness, :].set(0.0)
        pixels = pixels.at[-border_thickness:, :].set(0.0)

        im = Image.fromarray(np.array(pixels.astype(jnp.uint8)))
        im.save(f"results/screenshot_{str(mtime)}.png")

    def _get_selected_shape_global_indices(self):
        def _idx(idx, type):
            if type == ObjectType.CIRCLE:
                return idx + self.static_env_params.num_polygons
            return idx

        indices_to_use = jnp.array([_idx(idx, type) for idx, type in self.all_selected_shapes])
        return indices_to_use

    # flag1
    def _handle_events(self, do_dummy=False):
        pygame_widgets.update(self.pygame_events)
        if do_dummy or self.selected_shape_index < 0:
            gravity_main = self.all_widgets[None]["sldGravity"].getValue()

            def _set_single_global(state, gravity):
                return state.replace(
                    gravity=state.gravity.at[1].set(gravity),
                )

            env_state = _set_single_global(self.env_state, gravity_main)
            if not do_dummy:
                self.env_state = env_state

        if self.edit_shape_mode == EditMode.SELECT or do_dummy:  # is on the hand.
            if self.selected_shape_index < 0 and not do_dummy:
                return
            if do_dummy or self.selected_shape_type in [ObjectType.POLYGON, ObjectType.CIRCLE]:  # rigidbody
                shape_main = select_object(self.env_state, self.selected_shape_type, self.selected_shape_index)

                parent_container_main = (
                    self.env_state.circle if self.selected_shape_type == ObjectType.CIRCLE else self.env_state.polygon
                )

                new_density_main = self.all_widgets[self.selected_shape_type]["sldDensity"].getValue()
                fixated = self.all_widgets[self.selected_shape_type]["tglFixate"].getValue()

                fix_val = 0.0 if fixated else 1.0

                def _density_calcs(base, new_density):
                    inv_mass = jax.lax.select(
                        self.selected_shape_type == ObjectType.CIRCLE,
                        calc_inverse_mass_circle(base.radius, new_density),
                        calc_inverse_mass_polygon(base.vertices, base.n_vertices, self.static_env_params, new_density)[
                            0
                        ],
                    )
                    inv_inertia = jax.lax.select(
                        self.selected_shape_type == ObjectType.CIRCLE,
                        calc_inverse_inertia_circle(base.radius, new_density),
                        calc_inverse_inertia_polygon(
                            base.vertices, base.n_vertices, self.static_env_params, new_density
                        ),
                    )

                    return inv_mass, inv_inertia

                inv_mass_main, inv_inertia_main = _density_calcs(shape_main, new_density_main)

                friction_main = self.all_widgets[self.selected_shape_type]["sldFriction"].getValue()
                restitution = self.all_widgets[self.selected_shape_type]["sldRestitution"].getValue()

                position_main = jnp.array(
                    [
                        self.all_widgets[self.selected_shape_type]["sldPosition_X"].getValue(),
                        self.all_widgets[self.selected_shape_type]["sldPosition_Y"].getValue(),
                    ]
                )

                rotation_main = self.all_widgets[self.selected_shape_type]["sldRotation"].getValue()

                velocity_main = jnp.array(
                    [
                        self.all_widgets[self.selected_shape_type]["sldVelocity_X"].getValue(),
                        self.all_widgets[self.selected_shape_type]["sldVelocity_Y"].getValue(),
                    ]
                )

                angular_velocity_main = self.all_widgets[self.selected_shape_type]["sldAngular_Velocity"].getValue()

                # Circle stuff
                radius_main = None
                if self.selected_shape_type == ObjectType.CIRCLE:
                    radius_main = self.all_widgets[self.selected_shape_type]["sldRadius"].getValue()

                # Poly stuff
                vertices_main = None
                if self.selected_shape_type == ObjectType.POLYGON:
                    # Triangle
                    new_size_main = self.all_widgets[self.selected_shape_type]["sldSize"].getValue()

                    current_size = jnp.abs(self.env_state.polygon.vertices[self.selected_shape_index]).max()
                    scale_main = new_size_main / current_size

                    vertices_main = self.env_state.polygon.vertices[self.selected_shape_index] * scale_main

                def _set_single_state_rbody(
                    state,
                    parent_container,
                    density,
                    inv_mass,
                    inv_inertia,
                    friction,
                    position,
                    radius,
                    rotation,
                    velocity,
                    angular_velocity,
                    vertices,
                ):
                    new = {
                        "friction": parent_container.friction.at[self.selected_shape_index].set(friction),
                        "collision_mode": parent_container.collision_mode.at[self.selected_shape_index].set(
                            int(self.all_widgets[self.selected_shape_type]["sldCollidability"].getValue())
                        ),
                        "inverse_mass": parent_container.inverse_mass.at[self.selected_shape_index].set(
                            inv_mass * fix_val
                        ),
                        "inverse_inertia": parent_container.inverse_inertia.at[self.selected_shape_index].set(
                            inv_inertia * fix_val
                        ),
                        "position": parent_container.position.at[self.selected_shape_index].set(position),
                        "rotation": parent_container.rotation.at[self.selected_shape_index].set(rotation),
                        "velocity": parent_container.velocity.at[self.selected_shape_index].set(velocity),
                        "angular_velocity": parent_container.angular_velocity.at[self.selected_shape_index].set(
                            angular_velocity
                        ),
                        "restitution": parent_container.restitution.at[self.selected_shape_index].set(restitution),
                    }

                    if self.selected_shape_type == ObjectType.CIRCLE:
                        state = state.replace(
                            circle=state.circle.replace(
                                **new,
                                radius=parent_container.radius.at[self.selected_shape_index].set(radius),
                            ),
                            circle_shape_roles=state.circle_shape_roles.at[self.selected_shape_index].set(
                                int(self.all_widgets[self.selected_shape_type]["sldRole"].getValue())
                            ),
                            circle_densities=state.circle_densities.at[self.selected_shape_index].set(density),
                        )
                    else:
                        state = state.replace(
                            polygon=state.polygon.replace(
                                **new,
                                vertices=parent_container.vertices.at[self.selected_shape_index].set(vertices),
                            ),
                            polygon_shape_roles=state.polygon_shape_roles.at[self.selected_shape_index].set(
                                int(self.all_widgets[self.selected_shape_type]["sldRole"].getValue())
                            ),
                            polygon_densities=state.polygon_densities.at[self.selected_shape_index].set(density),
                        )

                    return state

                position_delta = position_main - shape_main.position
                env_state = _set_single_state_rbody(
                    self.env_state,
                    parent_container_main,
                    new_density_main,
                    inv_mass_main,
                    inv_inertia_main,
                    friction_main,
                    position_main,
                    radius_main,
                    rotation_main,
                    velocity_main,
                    angular_velocity_main,
                    vertices_main,
                )
                if not do_dummy:
                    self.env_state = env_state

            if do_dummy or self.selected_shape_type == ObjectType.JOINT:  # rjoint
                speed_main = self.all_widgets[ObjectType.JOINT]["sldSpeed"].getValue()

                power_main = self.all_widgets[ObjectType.JOINT]["sldPower"].getValue()

                motor_binding_val_min = int(self.all_widgets[ObjectType.JOINT]["sldColour"].getValue())

                auto_motor_val = self.all_widgets[ObjectType.JOINT]["tglAutoMotor"].getValue()
                joint_limits_val = self.all_widgets[ObjectType.JOINT]["tglJointLimits"].getValue()
                is_fixed_val = self.all_widgets[ObjectType.JOINT]["tglIsFixedJoint"].getValue()
                is_motor_on = self.all_widgets[ObjectType.JOINT]["tglIsMotorOn"].getValue()

                min_rot_val = jnp.deg2rad(self.all_widgets[ObjectType.JOINT]["sldMin_Rotation"].getValue())
                max_rot_val = jnp.deg2rad(self.all_widgets[ObjectType.JOINT]["sldMax_Rotation"].getValue())
                # ensure the min is less than the max
                min_rot_val, max_rot_val = min(min_rot_val, max_rot_val), max(min_rot_val, max_rot_val)

                def _set_single_state_joint(state, speed, power, colour):
                    state = state.replace(
                        joint=state.joint.replace(
                            motor_speed=state.joint.motor_speed.at[self.selected_shape_index].set(speed),
                            motor_power=state.joint.motor_power.at[self.selected_shape_index].set(power),
                            motor_has_joint_limits=state.joint.motor_has_joint_limits.at[self.selected_shape_index].set(
                                joint_limits_val
                            ),
                            min_rotation=state.joint.min_rotation.at[self.selected_shape_index].set(min_rot_val),
                            max_rotation=state.joint.max_rotation.at[self.selected_shape_index].set(max_rot_val),
                            is_fixed_joint=state.joint.is_fixed_joint.at[self.selected_shape_index].set(is_fixed_val),
                            motor_on=state.joint.motor_on.at[self.selected_shape_index].set(is_motor_on),
                        ),
                        motor_bindings=state.motor_bindings.at[self.selected_shape_index].set(colour),
                        motor_auto=state.motor_auto.at[self.selected_shape_index].set(auto_motor_val),
                    )
                    return state

                env_state = _set_single_state_joint(
                    self.env_state,
                    speed_main,
                    power_main,
                    motor_binding_val_min,
                )

                if not do_dummy:
                    self.env_state = env_state

            if do_dummy or self.selected_shape_type == ObjectType.THRUSTER:  # thruster
                power_main = self.all_widgets[ObjectType.THRUSTER]["sldPower"].getValue()

                def _set_single_state_thruster(state, power):
                    return state.replace(
                        thruster=state.thruster.replace(
                            power=state.thruster.power.at[self.selected_shape_index].set(power),
                        ),
                        thruster_bindings=state.thruster_bindings.at[self.selected_shape_index].set(
                            int(self.all_widgets[ObjectType.THRUSTER]["sldColour"].getValue())
                        ),
                    )

                env_state = _set_single_state_thruster(self.env_state, power_main)
                if not do_dummy:
                    self.env_state = env_state

    # flag2
    def _put_state_values_into_gui(self, env_state=None):
        def _set_toggle_val(toggle, val):
            if toggle.getValue() != val:
                toggle.toggle()

        def _enable_slider(slider):
            slider.enable()
            slider.colour = (200, 200, 200)
            slider.handleColour = (0, 0, 0)

        def _disable_slider(slider):
            slider.disable()
            slider.colour = (255, 0, 0)
            slider.handleColour = (255, 0, 0)

        if env_state is None:
            # state = self.edit_state
            raise ValueError

        def make_text(main_val):
            return f"{main_val:.2f}"

        # global ones
        self.all_widgets[None]["lblGravity"].setText(f"Gravity: {make_text(env_state.gravity[1])}")
        self.all_widgets[None]["sldGravity"].setValue(env_state.gravity[1])

        if self.edit_shape_mode != EditMode.SELECT or self.selected_shape_index < 0:
            return

        obj_main = select_object(env_state, self.selected_shape_type, self.selected_shape_index)

        if len(self.all_selected_shapes) > 1:
            assert False

        if self.selected_shape_type == ObjectType.JOINT:
            self.all_widgets[ObjectType.JOINT]["lblSpeed"].setText(f"Speed: {make_text(obj_main.motor_speed)}")
            self.all_widgets[ObjectType.JOINT]["sldSpeed"].setValue(obj_main.motor_speed)

            self.all_widgets[ObjectType.JOINT]["lblPower"].setText(f"Power: {make_text(obj_main.motor_power)}")
            self.all_widgets[ObjectType.JOINT]["sldPower"].setValue(obj_main.motor_power)

            self.all_widgets[ObjectType.JOINT]["lblColour"].setText(
                f"Colour: {env_state.motor_bindings[self.selected_shape_index]}"
            )
            self.all_widgets[ObjectType.JOINT]["sldColour"].setValue(
                env_state.motor_bindings[self.selected_shape_index]
            )

            self.all_widgets[ObjectType.JOINT]["lblJointLimits"].setText(
                f"Joint Limits: {obj_main.motor_has_joint_limits}"
            )
            widget_motor_has_joint_limits = self.all_widgets[ObjectType.JOINT]["tglJointLimits"].getValue()
            if obj_main.motor_has_joint_limits != widget_motor_has_joint_limits:  # Update the toggle
                self.all_widgets[ObjectType.JOINT]["tglJointLimits"].toggle()

            mini, maxi = jnp.rad2deg(obj_main.min_rotation), jnp.rad2deg(obj_main.max_rotation)
            self.all_widgets[ObjectType.JOINT]["lblMin_Rotation"].setText(f"Min Rot: {int(mini)}")
            self.all_widgets[ObjectType.JOINT]["sldMin_Rotation"].setValue(mini)
            self.all_widgets[ObjectType.JOINT]["lblMax_Rotation"].setText(f"Max Rot: {int(maxi)}")
            self.all_widgets[ObjectType.JOINT]["sldMax_Rotation"].setValue(maxi)

            if not obj_main.motor_has_joint_limits:
                for k in ["min_rotation", "max_rotation"]:
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].disable()
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].colour = (255, 0, 0)
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].handleColour = (255, 0, 0)
            else:
                for k in ["min_rotation", "max_rotation"]:
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].enable()
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].colour = (200, 200, 200)
                    self.all_widgets[self.selected_shape_type][f"sld{k.title()}"].handleColour = (0, 0, 0)

            self.all_widgets[ObjectType.JOINT]["lblAutoMotor"].setText(
                f"Auto: {env_state.motor_auto[self.selected_shape_index]}"
            )
            widget_is_auto_motor = self.all_widgets[ObjectType.JOINT]["tglAutoMotor"].getValue()
            if env_state.motor_auto[self.selected_shape_index] != widget_is_auto_motor:  # Update the toggle
                self.all_widgets[ObjectType.JOINT]["tglAutoMotor"].toggle()

            self.all_widgets[ObjectType.JOINT]["lblIsFixedJoint"].setText(f"Fixed: {obj_main.is_fixed_joint}")
            widget_is_motor_on = self.all_widgets[ObjectType.JOINT]["tglIsFixedJoint"].getValue()
            if obj_main.is_fixed_joint != widget_is_motor_on:  # Update the toggle
                self.all_widgets[ObjectType.JOINT]["tglIsFixedJoint"].toggle()

            self.all_widgets[ObjectType.JOINT]["lblIsMotorOn"].setText(f"Motor On: {obj_main.motor_on}")
            widget_is_motor_on = self.all_widgets[ObjectType.JOINT]["tglIsMotorOn"].getValue()
            if obj_main.motor_on != widget_is_motor_on:  # Update the toggle
                self.all_widgets[ObjectType.JOINT]["tglIsMotorOn"].toggle()

        elif self.selected_shape_type == ObjectType.THRUSTER:
            # thruster
            self.all_widgets[ObjectType.THRUSTER]["lblPower"].setText(f"Power: {make_text(obj_main.power)}")
            self.all_widgets[ObjectType.THRUSTER]["sldPower"].setValue(obj_main.power)
            self.all_widgets[ObjectType.THRUSTER]["sldColour"].setValue(
                env_state.thruster_bindings[self.selected_shape_index]
            )
            self.all_widgets[ObjectType.THRUSTER]["lblColour"].setText(
                f"Colour: {env_state.thruster_bindings[self.selected_shape_index]}"
            )

        elif self.selected_shape_type in [ObjectType.POLYGON, ObjectType.CIRCLE]:
            # rigidbody

            # Position
            # We use the mask for position_x for the entire position vector
            self.all_widgets[self.selected_shape_type]["lblPosition_X"].setText(
                f"Position X: {make_text(obj_main.position[0])}"
            )
            self.all_widgets[self.selected_shape_type]["sldPosition_X"].setValue(obj_main.position[0])

            self.all_widgets[self.selected_shape_type]["lblPosition_Y"].setText(
                f"Position Y: {make_text(obj_main.position[1])}"
            )
            self.all_widgets[self.selected_shape_type]["sldPosition_Y"].setValue(obj_main.position[1])

            # Velocity
            # We use the mask for velocity_x for the entire velocity vector
            self.all_widgets[self.selected_shape_type]["lblVelocity_X"].setText(
                f"Velocity X: {make_text(obj_main.velocity[0])}"
            )
            self.all_widgets[self.selected_shape_type]["sldVelocity_X"].setValue(obj_main.velocity[0])

            self.all_widgets[self.selected_shape_type]["lblVelocity_Y"].setText(
                f"Velocity Y: {make_text(obj_main.velocity[1])}"
            )
            self.all_widgets[self.selected_shape_type]["sldVelocity_Y"].setValue(obj_main.velocity[1])

            # Density
            is_fixated = obj_main.inverse_mass == 0

            def _calc_density(state):
                if self.selected_shape_type == ObjectType.POLYGON:
                    return state.polygon_densities[self.selected_shape_index]
                elif self.selected_shape_type == ObjectType.CIRCLE:
                    return state.circle_densities[self.selected_shape_index]
                else:
                    raise ValueError

            density_main = _calc_density(env_state)

            self.all_widgets[self.selected_shape_type]["lblDensity"].setText(f"Density: {make_text(density_main)}")
            self.all_widgets[self.selected_shape_type]["sldDensity"].setValue(density_main)
            if is_fixated:
                _disable_slider(self.all_widgets[self.selected_shape_type]["sldDensity"])
            else:
                _enable_slider(self.all_widgets[self.selected_shape_type]["sldDensity"])

            # Friction
            self.all_widgets[self.selected_shape_type]["lblFriction"].setText(
                f"Friction: {make_text(obj_main.friction)}"
            )
            self.all_widgets[self.selected_shape_type]["sldFriction"].setValue(obj_main.friction)

            # Restitution
            self.all_widgets[self.selected_shape_type]["lblRestitution"].setText(
                f"Restitution: {obj_main.restitution:.2f}"
            )
            self.all_widgets[self.selected_shape_type]["sldRestitution"].setValue(obj_main.restitution)

            # Rotation
            self.all_widgets[self.selected_shape_type]["lblRotation"].setText(
                f"Rotation: {make_text(obj_main.rotation)}"
            )
            self.all_widgets[self.selected_shape_type]["sldRotation"].setValue(obj_main.rotation)

            # Angular_Velocity
            self.all_widgets[self.selected_shape_type]["lblAngular_Velocity"].setText(
                f"Angular_Velocity: {make_text(obj_main.angular_velocity)}"
            )
            self.all_widgets[self.selected_shape_type]["sldAngular_Velocity"].setValue(obj_main.angular_velocity)

            # Collision mode
            self.all_widgets[self.selected_shape_type]["lblCollidability"].setText(
                f"Collidability: {obj_main.collision_mode}"
            )
            self.all_widgets[self.selected_shape_type]["sldCollidability"].setValue(obj_main.collision_mode)

            # Shape role
            if self.selected_shape_type == ObjectType.POLYGON:
                shape_role = env_state.polygon_shape_roles[self.selected_shape_index]
            else:
                shape_role = env_state.circle_shape_roles[self.selected_shape_index]

            self.all_widgets[self.selected_shape_type]["sldRole"].setValue(shape_role)
            self.all_widgets[self.selected_shape_type]["lblRole"].setText(f"Role: {shape_role}")

            # Fixate
            self.all_widgets[self.selected_shape_type]["lblFixate"].setText(f"Fixate: {is_fixated}")
            widget_is_fixed = self.all_widgets[self.selected_shape_type]["tglFixate"].getValue()
            if is_fixated != widget_is_fixed:  # Update the toggle
                self.all_widgets[self.selected_shape_type]["tglFixate"].toggle()

            # Radius
            if self.selected_shape_type == ObjectType.CIRCLE:
                self.all_widgets[self.selected_shape_type]["lblRadius"].setText(f"Radius: {make_text(obj_main.radius)}")
                self.all_widgets[self.selected_shape_type]["sldRadius"].setValue(obj_main.radius)

            elif self.selected_shape_type == ObjectType.POLYGON:
                size_main = jnp.abs(obj_main.vertices).max()

                self.all_widgets[self.selected_shape_type]["lblSize"].setText(f"Size: {make_text(size_main)}")
                self.all_widgets[self.selected_shape_type]["sldSize"].setValue(size_main)

    def _render_side_panel(self):
        arr = jnp.ones((self.side_panel_width, self.static_env_params.screen_dim[1], 3)) * (
            jnp.array([135.0, 206.0, 235.0])[None, None] + 20
        )
        return arr

    def make_label_and_slider(
        self,
        start_y,
        label_text,
        slider_min=0.0,
        slider_max=1.0,
        slider_step=0.05,
        font_size=18,
        is_toggle=False,
    ):
        Wl = round(self.W * 0.7)
        label = TextBox(
            self.screen_surface,
            self.MARGIN,
            start_y,
            Wl,
            20,
            fontSize=font_size,
            margin=0,
            placeholderText=label_text,
            font=pygame.font.SysFont("sans-serif", font_size),
        )
        label.disable()  # Act as label instead of textbox
        if is_toggle:
            widget = Toggle(self.screen_surface, self.W // 2 - 20, start_y + 23, 20, 13)
        else:
            widget = Slider(
                self.screen_surface,
                self.MARGIN,
                start_y + 23,
                Wl,
                13,
                min=slider_min,
                max=slider_max,
                step=slider_step,
            )

        return label, widget

    def _setup_side_panel(self):

        W = self.W = int(self.side_panel_width * self.upscale * 0.8)
        MARGIN = self.MARGIN = (self.side_panel_width * self.upscale - W) // 2

        # global values
        G = {}
        gravity_label, gravity_slider = self.make_label_and_slider(150, "Gravity", -20.0, 0.01, 0.2)
        G["lblGravity"] = gravity_label
        G["sldGravity"] = gravity_slider

        # thruster values
        T = {}
        thruster_power_label, thruster_power_slider = self.make_label_and_slider(150, "Power", slider_max=3.0)
        T["lblPower"] = thruster_power_label
        T["sldPower"] = thruster_power_slider

        thruster_colour_label, thruster_colour_slider = self.make_label_and_slider(
            250, "Colour", 0, self.static_env_params.num_thruster_bindings - 1, 1
        )
        T["lblColour"] = thruster_colour_label
        T["sldColour"] = thruster_colour_slider

        # joints
        D = {}
        for i, (name, (mini, maxi, step)) in enumerate(
            zip(
                ["speed", "power", "colour", "min_rotation", "max_rotation"],
                [
                    (-3, 3, 0.05),
                    (0, 3, 0.05),
                    (0, self.static_env_params.num_motor_bindings - 1, 1),
                    (-180, 180, 5),
                    (-180, 180, 5),
                ],
            )
        ):
            label, slider = self.make_label_and_slider(
                150 + 80 * i,
                f"Motor {name.title()}",
                slider_min=mini,
                slider_max=maxi,
                slider_step=step,
            )
            D["lbl" + name.title()] = label
            D["sld" + name.title()] = slider

        label, toggle = self.make_label_and_slider(150 + 80 * 5, "Joint Limits", is_toggle=True)
        D["lblJointLimits"] = label
        D["tglJointLimits"] = toggle

        label, toggle = self.make_label_and_slider(150 + 80 * 6, "Auto", is_toggle=True)
        D["lblAutoMotor"] = label
        D["tglAutoMotor"] = toggle

        label, toggle = self.make_label_and_slider(150 + 80 * 7, "Fixed", is_toggle=True)
        D["lblIsFixedJoint"] = label
        D["tglIsFixedJoint"] = toggle

        label, toggle = self.make_label_and_slider(150 + 80 * 8, "Motor On", is_toggle=True)
        D["lblIsMotorOn"] = label
        D["tglIsMotorOn"] = toggle

        def _create_rigid_body_base_gui():
            D_rigid = {}
            # rigidbodies
            total_toggles = 0
            total_non_toggles = 0
            for i, (name, bounds) in enumerate(
                zip(
                    [
                        "position_x",
                        "position_y",
                        "rotation",
                        "velocity_x",
                        "velocity_y",
                        "angular_velocity",
                        "density",
                        "friction",
                        "restitution",
                        "collidability",
                        "role",
                    ],
                    [
                        (0, 5.0, 0.01),
                        (0, 5.0, 0.01),
                        (-2 * jnp.pi, 2 * jnp.pi, 0.01),
                        (-10.0, 10.0, 0.1),
                        (-10.0, 10.0, 0.1),
                        (-6, 6.0, 0.01),
                        (0.1, 5.0, 0.1),
                        (0.02, 1.0, 0.02),
                        (0.0, 0.8, 0.02),
                        (0, 2, 1),
                        (0, 3, 1),
                    ],
                )
            ):
                location = 50 + 80 * total_toggles + 40 * total_non_toggles
                label, slider = self.make_label_and_slider(
                    location,
                    name.title(),
                    *bounds,
                )
                total_toggles += 0
                total_non_toggles += 1
                D_rigid["lbl" + name.title()] = label
                D_rigid["sld" + name.title()] = slider

            location = 50 + 80 * total_toggles + 40 * total_non_toggles
            # toggles:
            label, toggle = self.make_label_and_slider(location, "Fixate", is_toggle=True)
            D_rigid["lblFixate"] = label
            D_rigid["tglFixate"] = toggle

            return D_rigid, location

        # Circle extras
        D_circle, location = _create_rigid_body_base_gui()
        label, slider = self.make_label_and_slider(
            location + 40,
            "Radius",
            slider_min=0.1,
            slider_max=1.0,
            slider_step=0.02,
        )
        D_circle["lblRadius"] = label
        D_circle["sldRadius"] = slider

        # Polygon extras
        D_poly, location = _create_rigid_body_base_gui()
        label, slider = self.make_label_and_slider(
            location + 40,
            "Size",
            slider_min=0.1,
            slider_max=2.0,
            slider_step=0.02,
        )
        D_poly["lblSize"] = label
        D_poly["sldSize"] = slider

        self.all_widgets = {
            ObjectType.THRUSTER: T,
            ObjectType.JOINT: D,
            "GENERAL": {
                "lblGeneral": TextBox(
                    self.screen_surface,
                    MARGIN,
                    10,
                    W,
                    30,
                    fontSize=20,
                    margin=0,
                    placeholderText="General",
                    font=pygame.font.SysFont("sans-serif", 35),
                ),
            },
            ObjectType.POLYGON: D_poly,
            ObjectType.CIRCLE: D_circle,
            None: G,
        }

        self._hide_all_widgets()

    def _render_edit_overlay(self, pixels, is_editing, edit_shape_mode):
        is_editing_texture = jax.lax.select(is_editing, EDIT_TEXTURE_RGBA, PLAY_TEXTURE_RGBA)
        is_editing_texture = jnp.repeat(jnp.repeat(is_editing_texture, self.upscale, axis=0), self.upscale, axis=1)

        offset = self.side_panel_width * self.upscale
        w = 64 * self.upscale
        offset2 = int(w * 1.25)
        offset_y = 16 * self.upscale

        play_tex_with_background = (1 - is_editing_texture[:, :, 3:]) * pixels[
            offset + 0 : offset + w, 0:w
        ] + is_editing_texture[:, :, 3:] * is_editing_texture[:, :, :3]
        pixels = pixels.at[offset : offset + w, 0:w].set(play_tex_with_background)

        edit_shape_texture = jax.lax.switch(
            edit_shape_mode,
            [
                lambda: CIRCLE_TEXTURE_RGBA,
                lambda: RECT_TEXTURE_RGBA,
                lambda: RJOINT_TEXTURE_RGBA,
                lambda: SELECT_TEXTURE_RGBA,
                lambda: TRIANGLE_TEXTURE_RGBA,
                lambda: THRUSTER_TEXTURE_RGBA,
            ],
        )

        edit_shape_texture = jnp.repeat(jnp.repeat(edit_shape_texture, self.upscale, axis=0), self.upscale, axis=1)

        edit_shape_texture_alpha = edit_shape_texture[:, :, 3:] * is_editing

        w = 32 * self.upscale
        edit_shape_texture_with_background = (1 - edit_shape_texture_alpha) * pixels[
            offset + offset2 : offset + offset2 + w, offset_y : offset_y + w
        ] + edit_shape_texture_alpha * edit_shape_texture[:, :, :3]
        pixels = pixels.at[offset + offset2 : offset + offset2 + w, offset_y : offset_y + w].set(
            edit_shape_texture_with_background
        )

        return pixels

    def edit(self):
        self.rng, _rng = jax.random.split(self.rng)
        env_state = self.env_state

        left_click = False
        right_click = False

        keys = []
        keys_up_this_frame = set()
        for event in self.pygame_events:
            if event.type == pygame.KEYDOWN:
                if (
                    event.key == pygame.K_s
                    and (pygame.key.get_mods() & pygame.KMOD_CTRL)
                    and (pygame.key.get_mods() & pygame.KMOD_SHIFT)
                ):
                    filename = prompt_file(save=True)
                    if filename:
                        filename += ".level.pkl"
                        save_pickle(filename, self.last_played_level)
                        print(f"Saved last sampled level to {filename}")
                elif event.key == pygame.K_s and (pygame.key.get_mods() & pygame.KMOD_CTRL):
                    env_state = self._reset_select_shape(env_state)
                    filename = prompt_file(save=True)
                    if filename:
                        if filename.endswith(".json"):
                            export_env_state_to_json(filename, env_state, self.static_env_params, self.env_params)
                        elif not filename.endswith(".pcg.pkl"):
                            filename += ".pcg.pkl"
                            save_pickle(filename, env_state)
                        print(f"Saved PCG state to {filename}")
                elif event.key == pygame.K_o and (pygame.key.get_mods() & pygame.KMOD_CTRL):
                    filename = prompt_file(save=False)
                    if filename:
                        self._reset_select_shape(env_state)
                        if filename.endswith(".json"):
                            env_state, new_static_env_params, new_env_params = load_from_json_file(filename)
                            self._update_params(new_static_env_params, new_env_params, env_state=env_state)
                    self._reset_triangles()
                elif event.key == pygame.K_n and (pygame.key.get_mods() & pygame.KMOD_CTRL):
                    self._reset_select_shape(env_state)
                    env_state = new_env(self.static_env_params)
                    self._reset_triangles()
                else:
                    keys.append(event.key)
            elif event.type == pygame.KEYUP:
                keys_up_this_frame.add(event.key)
            if event.type == pygame.MOUSEBUTTONDOWN and self._get_mouse_position_world_space()[0] >= 0:
                if event.button == 1:
                    left_click = True
                if event.button == 3:
                    right_click = True

            if event.type == pygame.MOUSEWHEEL:
                env_state = self._handle_scroll_wheel(env_state, event.y)

        if self.selected_shape_index == -1:
            num = get_numeric_key_pressed(self.pygame_events)
            if num is not None:
                self.edit_shape_mode = EditMode(num % len(EditMode))

        # We have to do these checks outside the loop, otherwise they get triggered multiple times per key press.
        if pygame.key.get_mods() & pygame.KMOD_SHIFT:
            state = env_state
            if pygame.K_m in keys_up_this_frame:
                state = self.mutate_world(_rng, state, 1)
            if pygame.K_c in keys_up_this_frame:
                state, _ = mutate_add_connected_shape(
                    _rng, state, self.env_params, self.static_env_params, self.ued_params
                )
            elif pygame.K_s in keys_up_this_frame:
                state = mutate_add_shape(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_p in keys_up_this_frame:
                state = mutate_swap_role(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_r in keys_up_this_frame:
                state = mutate_remove_shape(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_j in keys_up_this_frame:
                state = mutate_remove_joint(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_t in keys_up_this_frame:
                state = mutate_toggle_fixture(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_g in keys_up_this_frame:
                state = mutate_add_thruster(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_l in keys_up_this_frame:
                state = mutate_remove_thruster(_rng, state, self.env_params, self.static_env_params, self.ued_params)
            elif pygame.K_b in keys_up_this_frame:
                state = self.mutate_change_shape_size(
                    _rng, state, self.env_params, self.static_env_params, self.ued_params
                )
            elif pygame.K_x in keys_up_this_frame:
                state = mutate_change_shape_location(
                    _rng, state, self.env_params, self.static_env_params, self.ued_params
                )
            elif pygame.K_k in keys_up_this_frame:
                state = self.mutate_change_shape_rotation(
                    _rng, state, self.env_params, self.static_env_params, self.ued_params
                )
            env_state = state

        if pygame.K_p in keys_up_this_frame:
            global myrng
            myrng, _rng = jax.random.split(myrng)
            # use the same rng
            env_state = permute_state(_rng, env_state, self.static_env_params)
        if self.edit_shape_mode == EditMode.SELECT:  # select a shape
            env_state = self._edit_select_shape(env_state, left_click, right_click, keys)
            self._put_state_values_into_gui(env_state)
            env_state = self._select_shape_keyboard_shortcuts(env_state, left_click, keys)
        else:
            env_state = self._reset_select_shape(env_state)  # don't highlight
            self._put_state_values_into_gui(env_state)
            self._show_correct_widgets(None)

        if self.edit_shape_mode != EditMode.ADD_TRIANGLE or not self.creating_shape:
            self.num_triangle_clicks = 0

        if self.edit_shape_mode == EditMode.ADD_CIRCLE:
            env_state = self._edit_circle(env_state, left_click, right_click)
        elif self.edit_shape_mode == EditMode.ADD_RECTANGLE:
            env_state = self._edit_rect(env_state, left_click, right_click)
        elif self.edit_shape_mode == EditMode.ADD_JOINT:
            env_state = self._edit_joint(env_state, left_click, right_click)
        elif self.edit_shape_mode == EditMode.ADD_TRIANGLE:
            env_state = self._edit_triangle(env_state, left_click, right_click)
        elif self.edit_shape_mode == EditMode.ADD_THRUSTER:
            env_state = self._edit_thruster(env_state, left_click, right_click)

        env_state = recompute_global_joint_positions(
            env_state.replace(
                collision_matrix=calculate_collision_matrix(self.static_env_params, env_state.joint),
            ),
            self.static_env_params,
        )
        return env_state

    def _update_params(
        self, new_static_env_params: StaticEnvParams, new_env_params: EnvParams, env_state: EnvState = None
    ):
        self.static_env_params = new_static_env_params.replace(
            frame_skip=self.config["frame_skip"], downscale=self.config["downscale"]
        )
        self.env_params = new_env_params
        self.env = make_kinetix_env(
            ActionType.MULTI_DISCRETE,
            ObservationType.SYMBOLIC_ENTITY,
            make_reset_function(self.static_env_params),
            self.env_params,
            self.static_env_params,
        )
        self._setup_rendering(self.static_env_params, self.env_params)
        self._step_fn = jax.jit(self.env.step)
        self._jit_step(env_state)  # jit so that it doesn't take a long time when pressing play

    def _discard_shape_being_created(self, env_state):
        if self.creating_shape:
            if self.edit_shape_mode == EditMode.ADD_CIRCLE:
                env_state = env_state.replace(
                    circle=env_state.circle.replace(
                        active=env_state.circle.active.at[self.creating_shape_index].set(False)
                    )
                )

            elif self.edit_shape_mode == EditMode.ADD_RECTANGLE:
                env_state = env_state.replace(
                    polygon=env_state.polygon.replace(
                        active=env_state.polygon.active.at[self.creating_shape_index].set(False)
                    )
                )

        self.creating_shape = False

        return env_state

    def _handle_scroll_wheel(self, env_state, y):
        if y == 0:
            return env_state

        state = self._discard_shape_being_created(env_state)

        self.edit_shape_mode = EditMode((self.edit_shape_mode.value + y) % len(EditMode))

        return state

    def _get_mouse_position_world_space(self):
        mouse_pos = pygame.mouse.get_pos()
        return (
            jnp.array(
                [
                    mouse_pos[0] / self.upscale - self.side_panel_width,
                    self.static_env_params.screen_dim[1] - mouse_pos[1] / self.upscale,
                ]
            )
            / self.env_params.pixels_per_unit
        )

    def _get_circles_on_mouse(self, state):
        mouse_pos = self._get_mouse_position_world_space()

        cis = []

        for ci in jnp.arange(self.static_env_params.num_circles)[::-1]:
            circle = jax.tree.map(lambda x: x[ci], state.circle)

            if not circle.active:
                continue

            dist = jnp.linalg.norm(mouse_pos - circle.position)
            if dist <= circle.radius:
                cis.append(ci)

        return cis

    def _get_revolute_joints_on_mouse(self, state: EnvState):
        mouse_pos = self._get_mouse_position_world_space()

        ris = []

        for ri in jnp.arange(self.static_env_params.num_joints)[::-1]:
            joint = jax.tree.map(lambda x: x[ri], state.joint)

            if not joint.active:
                continue

            dist = jnp.linalg.norm(mouse_pos - joint.global_position)
            if dist <= 10 / 100:  # arbitrary
                ris.append(ri)

        return ris

    def _get_thrusters_on_mouse(self, state: EnvState):
        mouse_pos = self._get_mouse_position_world_space()

        ris = []

        for ri in jnp.arange(self.static_env_params.num_thrusters)[::-1]:
            thruster = jax.tree.map(lambda x: x[ri], state.thruster)

            if not thruster.active:
                continue

            dist = jnp.linalg.norm(mouse_pos - thruster.global_position)
            if dist <= 16 / 100:  # arbitrary
                ris.append(ri)

        return ris

    def _get_joints_attached_to_shape(self, state, shape_index):
        r_a = jnp.arange(self.static_env_params.num_joints)[state.joint.a_index == shape_index]
        r_b = jnp.arange(self.static_env_params.num_joints)[state.joint.b_index == shape_index]

        t = jnp.arange(self.static_env_params.num_thrusters)[state.thruster.object_index == shape_index]
        return jnp.concatenate([r_a, r_b], axis=0), t

    def _edit_thruster(self, env_state: EnvState, left_click: bool, right_click: bool):
        if not self.creating_shape and (1 - env_state.thruster.active.astype(int)).sum() == 0:
            if not right_click:
                return env_state

        thruster_pos = self._get_mouse_position_world_space()
        idx = -1
        for ri in self._get_polygons_on_mouse(env_state):
            r = jax.tree.map(lambda x: x[ri], env_state.polygon)
            thruster_pos = snap_to_polygon_center_line(r, thruster_pos)
            relative_pos = jnp.matmul(rmat(r.rotation).transpose((1, 0)), thruster_pos - r.position)
            idx = ri
            break
        if idx == -1:
            for ci in self._get_circles_on_mouse(env_state):
                c = jax.tree.map(lambda x: x[ci], env_state.circle)
                thruster_pos = snap_to_center(c, thruster_pos)
                thruster_pos = snap_to_circle_center_line(c, thruster_pos)
                relative_pos = thruster_pos - c.position
                idx = ci + self.static_env_params.num_polygons
                break
        if left_click:
            if self.creating_shape:
                self.creating_shape = False
            else:
                if idx >= 0:
                    self.creating_shape = True
                    self.creating_shape_position = thruster_pos
                    self.creating_shape_index = jnp.argmin(env_state.thruster.active)
                    shape = select_shape(env_state, idx, self.static_env_params)

                    def _add_thruster_to_state(state):
                        state = state.replace(
                            thruster=state.thruster.replace(
                                object_index=state.thruster.object_index.at[self.creating_shape_index].set(idx),
                                relative_position=state.thruster.relative_position.at[self.creating_shape_index].set(
                                    relative_pos
                                ),
                                power=state.thruster.power.at[self.creating_shape_index].set(
                                    1.0 / jax.lax.select(shape.inverse_mass == 0, 1.0, shape.inverse_mass)
                                ),
                                active=state.thruster.active.at[self.creating_shape_index].set(True),
                                global_position=state.thruster.global_position.at[self.creating_shape_index].set(
                                    thruster_pos
                                ),
                                rotation=state.thruster.rotation.at[self.creating_shape_index].set(0.0),
                            ),
                            thruster_bindings=state.thruster_bindings.at[self.creating_shape_index].set(0),
                        )
                        return state

                    env_state = _add_thruster_to_state(env_state)
        elif right_click:
            for ti in self._get_thrusters_on_mouse(env_state):

                def _remove_thruster_from_state(state):
                    return state.replace(
                        thruster=state.thruster.replace(active=state.thruster.active.at[ti].set(False))
                    )

                return _remove_thruster_from_state(env_state)
        else:
            if self.creating_shape:
                curr_pos = self._get_mouse_position_world_space()

                normal = env_state.thruster.relative_position[self.creating_shape_index]
                angle = jnp.arctan2(normal[1], normal[0])

                relative_pos = curr_pos - self.creating_shape_position
                # rotation = jnp.arctan2(relative_pos[1], relative_pos[0])
                rotation = jnp.pi + jnp.arctan2(relative_pos[1], relative_pos[0]) + angle
                angle_round = jnp.round(rotation / (jnp.pi / 2))
                angle_norm = rotation / (jnp.pi / 2)
                if jnp.abs(angle_round - angle_norm) < 0.3:
                    rotation = angle_round * (jnp.pi / 2)

                def _update_thruster_rotation(state):
                    return state.replace(
                        thruster=state.thruster.replace(
                            rotation=state.thruster.rotation.at[self.creating_shape_index].set(rotation - angle),
                        )
                    )

                env_state = _update_thruster_rotation(env_state)
            else:
                pass

        return env_state

    def _edit_circle(self, env_state: EnvState, left_click: bool, right_click: bool):
        if right_click:
            for ci in self._get_circles_on_mouse(env_state):
                attached_j, attached_t = self._get_joints_attached_to_shape(
                    env_state, ci + self.static_env_params.num_polygons
                )

                def _remove_circle_from_state(state):
                    return state.replace(
                        circle=state.circle.replace(active=state.circle.active.at[ci].set(False)),
                        joint=state.joint.replace(active=state.joint.active.at[attached_j].set(False)),
                        thruster=state.thruster.replace(active=state.thruster.active.at[attached_t].set(False)),
                    )

                env_state = _remove_circle_from_state(env_state)
                env_state = env_state.replace(
                    collision_matrix=calculate_collision_matrix(self.static_env_params, env_state.joint)
                )

                return env_state

        if not self.creating_shape and (1 - env_state.circle.active.astype(int)).sum() == 0:
            return env_state

        radius = jnp.linalg.norm(self._get_mouse_position_world_space() - self.create_shape_position)
        radius = jnp.clip(radius, 5.0 / self.env_params.pixels_per_unit, self.static_env_params.max_shape_size / 2)

        def _add_circle(state, highlight):
            state = state.replace(
                circle=state.circle.replace(
                    position=state.circle.position.at[self.creating_shape_index].set(self.create_shape_position),
                    velocity=state.circle.velocity.at[self.creating_shape_index].set(jnp.array([0.0, 0.0])),
                    radius=state.circle.radius.at[self.creating_shape_index].set(radius),
                    inverse_mass=state.circle.inverse_mass.at[self.creating_shape_index].set(1.0),
                    inverse_inertia=state.circle.inverse_inertia.at[self.creating_shape_index].set(1.0),
                    active=state.circle.active.at[self.creating_shape_index].set(True),
                    collision_mode=state.circle.collision_mode.at[self.creating_shape_index].set(1),
                ),
                circle_shape_roles=state.circle_shape_roles.at[self.creating_shape_index].set(0),
                circle_highlighted=state.circle_highlighted.at[self.creating_shape_index].set(highlight),
                circle_densities=state.circle_densities.at[self.creating_shape_index].set(1.0),
            )
            return state

        if left_click:
            if self.creating_shape:
                env_state = _add_circle(env_state, False)

                env_state = recalculate_mass_and_inertia(
                    env_state,
                    self.static_env_params,
                    env_state.polygon_densities,
                    env_state.circle_densities,
                )

                self.creating_shape = False
            else:
                self.creating_shape_index = jnp.argmin(env_state.circle.active)
                self.create_shape_position = self._get_mouse_position_world_space()
                self.creating_shape = True

        else:
            if self.creating_shape:
                env_state = _add_circle(env_state, True)
            else:
                pass

        return env_state

    def _get_polygons_on_mouse(self, state, n_vertices=None):
        # n_vertices=None selects both triangles and quads
        ris = []

        mouse_pos = self._get_mouse_position_world_space()

        for ri in jnp.arange(self.static_env_params.num_polygons)[::-1]:
            polygon = jax.tree.map(lambda x: x[ri], state.polygon)

            if (not polygon.active) or ((n_vertices is not None) and polygon.n_vertices != n_vertices):
                continue

            mpos = rmat(-polygon.rotation) @ (mouse_pos - polygon.position)

            ii = jnp.arange(polygon.n_vertices)
            ii2 = (ii + 1) % polygon.n_vertices
            distances = jax.vmap(_mouse_dist_to_vertices, (None, 0, 0, None))(mpos, ii, ii2, polygon.vertices)
            if not (distances > 0).any():
                ris.append(ri)
        return ris

    def _edit_rect(self, env_state: EnvState, left_click: bool, right_click: bool):
        if right_click:
            for ri in self._get_polygons_on_mouse(env_state, n_vertices=4):
                attached_j, attached_t = self._get_joints_attached_to_shape(env_state, ri)

                def _remove_rect_from_state(state):
                    state = state.replace(
                        polygon=state.polygon.replace(
                            active=state.polygon.active.at[ri].set(False),
                            rotation=state.polygon.rotation.at[ri].set(0.0),
                        ),
                        joint=state.joint.replace(active=state.joint.active.at[attached_j].set(False)),
                        thruster=state.thruster.replace(active=state.thruster.active.at[attached_t].set(False)),
                    )

                    return state

                env_state = _remove_rect_from_state(env_state)

                env_state = env_state.replace(
                    collision_matrix=calculate_collision_matrix(self.static_env_params, env_state.joint)
                )

                return env_state

        if not self.creating_shape and (1 - env_state.polygon.active.astype(int)).sum() == 0:
            return env_state

        diff = (self._get_mouse_position_world_space() - self.create_shape_position) / 2
        diff = jnp.clip(
            diff,
            -(self.static_env_params.max_shape_size / 2) / jnp.sqrt(2),
            (self.static_env_params.max_shape_size / 2) / jnp.sqrt(2),
        )
        half_dim = jnp.abs(diff)
        half_dim = jnp.clip(half_dim, min=5.0 / self.env_params.pixels_per_unit)
        vertices = rectangle_vertices(half_dim)

        def _add_rect_to_state(state, highlight):
            state = state.replace(
                polygon=state.polygon.replace(
                    position=state.polygon.position.at[self.creating_shape_index].set(
                        self.create_shape_position + diff
                    ),
                    velocity=state.polygon.velocity.at[self.creating_shape_index].set(jnp.array([0.0, 0.0])),
                    vertices=state.polygon.vertices.at[self.creating_shape_index].set(vertices),
                    inverse_mass=state.polygon.inverse_mass.at[self.creating_shape_index].set(1.0),
                    inverse_inertia=state.polygon.inverse_inertia.at[self.creating_shape_index].set(1.0),
                    active=state.polygon.active.at[self.creating_shape_index].set(True),
                    collision_mode=state.polygon.collision_mode.at[self.creating_shape_index].set(1),
                    n_vertices=state.polygon.n_vertices.at[self.creating_shape_index].set(4),
                ),
                polygon_shape_roles=state.polygon_shape_roles.at[self.creating_shape_index].set(0),
                polygon_highlighted=state.polygon_highlighted.at[self.creating_shape_index].set(highlight),
                polygon_densities=state.polygon_densities.at[self.creating_shape_index].set(1.0),
            )

            return state

        if left_click:
            if self.creating_shape:
                env_state = _add_rect_to_state(env_state, False)

                env_state = recalculate_mass_and_inertia(
                    env_state,
                    self.static_env_params,
                    env_state.polygon_densities,
                    env_state.circle_densities,
                )

                self.creating_shape = False
            else:
                self.creating_shape_index = jnp.argmin(env_state.polygon.active)
                self.create_shape_position = self._get_mouse_position_world_space()
                self.creating_shape = True
        else:
            if self.creating_shape:
                env_state = _add_rect_to_state(env_state, True)

        return env_state

    def _reset_triangles(self):
        self.triangle_order = jnp.array([0, 1, 2])
        self.num_triangle_clicks = 0
        self.creating_shape = False

    def _edit_triangle(self, env_state: EnvState, left_click: bool, right_click: bool):
        if right_click:
            self.num_triangle_clicks = 0
            for ri in self._get_polygons_on_mouse(env_state, n_vertices=3):
                attached_r, attached_f = self._get_joints_attached_to_shape(env_state, ri)

                def _remove_triangle_from_state(state):
                    state = state.replace(
                        polygon=state.polygon.replace(
                            active=state.polygon.active.at[ri].set(False),
                            rotation=state.polygon.rotation.at[ri].set(0.0),
                        ),
                        joint=state.joint.replace(active=state.joint.active.at[attached_r].set(False)),
                    )

                    return state

                env_state = _remove_triangle_from_state(env_state)

                env_state = env_state.replace(
                    collision_matrix=calculate_collision_matrix(self.static_env_params, env_state.joint)
                )

                return env_state

        if not self.creating_shape and (1 - env_state.polygon.active.astype(int)).sum() == 0:
            return env_state

        def get_correct_center_two_verts(verts):
            return (jnp.max(verts, axis=0) + jnp.min(verts, axis=0)) / 2

        def order_clockwise(verts, loose_ordering=False):
            # verts has shape (3, 2), order them clockwise.
            # https://stackoverflow.com/questions/51074984/sort-vertices-in-clockwise-order
            # Calculate centroid
            centroid = jnp.mean(verts, axis=0)
            # Calculate angles
            angles = jnp.round(jnp.arctan2(verts[:, 1] - centroid[1], verts[:, 0] - centroid[0]), 2)
            # Order vertices
            order = jnp.argsort(-angles, stable=True)
            if loose_ordering:
                order = jnp.arange(len(order))
            ans = verts[order]
            # order is of shape (2, ) or (3, ). I want it to always be of shape 3
            if len(order) < 3:
                order = jnp.concatenate([order, jnp.array([2])])
            return ans, order

        def do_triangle_n_click(env_state, how_many_clicks, is_on_a_click=False):
            n = how_many_clicks
            # if we must keep them clockwise all the time, then the one we edit / move around may have varying indices.
            current_index_to_change = self.triangle_order[n]
            sign = 1
            idxs = jnp.arange(n + 1)
            idxs_to_allow = idxs[~(idxs == current_index_to_change)]

            # Get the new vertex and clip its position
            new_tentative_vert = (
                self._get_mouse_position_world_space() - env_state.polygon.position[self.creating_shape_index]
            )
            new_tentative_vert = jnp.clip(
                new_tentative_vert,
                jnp.max(env_state.polygon.vertices[self.creating_shape_index, idxs_to_allow], axis=0)
                - self.static_env_params.max_shape_size * 0.8,
                jnp.min(env_state.polygon.vertices[self.creating_shape_index, idxs_to_allow], axis=0)
                + self.static_env_params.max_shape_size * 0.8,
            )
            new_verts = env_state.polygon.vertices.at[self.creating_shape_index, current_index_to_change].set(
                new_tentative_vert
            )
            new_center_two = get_correct_center_two_verts(new_verts[self.creating_shape_index, : n + 1])

            _, new_center_three = calc_inverse_mass_polygon(
                new_verts[self.creating_shape_index],
                3,
                self.static_env_params,
                1.0,
            )
            new_center = jax.lax.select(n == 1, new_center_two, new_center_three)

            new_verts = new_verts.at[self.creating_shape_index].add(-sign * new_center)
            vvs = new_verts[self.creating_shape_index, : n + 1]

            ordered_vertices, new_permutation = order_clockwise(vvs, loose_ordering=not is_on_a_click)
            self.triangle_order = self.triangle_order[new_permutation]
            new_verts = new_verts.at[self.creating_shape_index, : n + 1].set(ordered_vertices)

            env_state = env_state.replace(
                polygon=env_state.polygon.replace(
                    vertices=new_verts,
                    position=env_state.polygon.position.at[self.creating_shape_index].add(sign * new_center),
                    n_vertices=env_state.polygon.n_vertices.at[self.creating_shape_index].set(n + 1),
                ),
            )

            return env_state

        if left_click:
            if self.creating_shape:
                assert 3 > self.num_triangle_clicks > 0
                if self.num_triangle_clicks == 1:
                    env_state = do_triangle_n_click(env_state, 1, is_on_a_click=True)
                    self.num_triangle_clicks += 1
                else:  # this finishes it
                    env_state = do_triangle_n_click(env_state, 2, is_on_a_click=True)
                    self.creating_shape = False
                    self.num_triangle_clicks = 0
                    env_state = recalculate_mass_and_inertia(
                        env_state,
                        self.static_env_params,
                        env_state.polygon_densities,
                        env_state.circle_densities,
                    )

            else:
                self.triangle_order = jnp.array([0, 1, 2])
                self.creating_shape_index = jnp.argmin(env_state.polygon.active)
                self.create_shape_position = self._get_mouse_position_world_space()
                self.creating_shape = True
                self.num_triangle_clicks = 1
                vertices = jnp.zeros((self.static_env_params.max_polygon_vertices, 2), dtype=jnp.float32)

                def _add_triangle_to_state(state):
                    state = state.replace(
                        polygon=state.polygon.replace(
                            position=state.polygon.position.at[self.creating_shape_index].set(
                                self.create_shape_position
                            ),
                            velocity=state.polygon.velocity.at[self.creating_shape_index].set(jnp.array([0.0, 0.0])),
                            vertices=state.polygon.vertices.at[self.creating_shape_index].set(vertices),
                            inverse_mass=state.polygon.inverse_mass.at[self.creating_shape_index].set(1.0),
                            inverse_inertia=state.polygon.inverse_inertia.at[self.creating_shape_index].set(1.0),
                            active=state.polygon.active.at[self.creating_shape_index].set(True),
                            n_vertices=state.polygon.n_vertices.at[self.creating_shape_index].set(1),
                        ),
                        polygon_shape_roles=state.polygon_shape_roles.at[self.creating_shape_index].set(0),
                        polygon_highlighted=state.polygon_highlighted.at[self.creating_shape_index].set(False),
                        polygon_densities=state.polygon_densities.at[self.creating_shape_index].set(1.0),
                    )

                    return state

                env_state = _add_triangle_to_state(env_state)

        elif self.creating_shape:
            assert 1 <= self.num_triangle_clicks <= 2
            env_state = do_triangle_n_click(
                env_state, self.num_triangle_clicks, is_on_a_click=self.num_triangle_clicks == 1
            )

        return env_state

    def _edit_joint(self, env_state: EnvState, left_click: bool, right_click: bool):
        if left_click and env_state.joint.active.all():
            return env_state

        if left_click:
            joint_index = jnp.argmin(env_state.joint.active)
            joint_position = self._get_mouse_position_world_space()
            # reverse them so that the joint order and rendering order remains the same.
            # We want the first shape to have a lower index than the second shape, with circles always having higher indices compared to rectangles.
            circles = self._get_circles_on_mouse(env_state)[::-1]
            rects = self._get_polygons_on_mouse(env_state)[::-1]

            if len(rects) + len(circles) >= 2:
                r1 = len(rects) >= 1
                r2 = len(rects) >= 2

                a_index = rects[0] if r1 else circles[0]  # + self.static_env_params.num_polygons
                b_index = rects[r1 * 1] if r2 else circles[1 - 1 * r1]  # + self.static_env_params.num_polygons

                a_shape = env_state.polygon if r1 else env_state.circle
                b_shape = env_state.polygon if r2 else env_state.circle

                a = jax.tree.map(lambda x: x[a_index], a_shape)
                b = jax.tree.map(lambda x: x[b_index], b_shape)

                a_index += (not r1) * self.static_env_params.num_polygons
                b_index += (not r2) * self.static_env_params.num_polygons

                joint_position = snap_to_center(a, joint_position)
                joint_position = snap_to_center(b, joint_position)

                a_relative_pos = jnp.matmul(rmat(a.rotation).transpose((1, 0)), joint_position - a.position)
                b_relative_pos = jnp.matmul(rmat(b.rotation).transpose((1, 0)), joint_position - b.position)

                def _add_joint_to_state(state):
                    state = state.replace(
                        joint=state.joint.replace(
                            a_index=state.joint.a_index.at[joint_index].set(a_index),
                            b_index=state.joint.b_index.at[joint_index].set(b_index),
                            a_relative_pos=state.joint.a_relative_pos.at[joint_index].set(a_relative_pos),
                            b_relative_pos=state.joint.b_relative_pos.at[joint_index].set(b_relative_pos),
                            active=state.joint.active.at[joint_index].set(True),
                            global_position=state.joint.global_position.at[joint_index].set(joint_position),
                            motor_on=state.joint.motor_on.at[joint_index].set(True),
                            motor_speed=state.joint.motor_speed.at[joint_index].set(1.0),
                            motor_power=state.joint.motor_power.at[joint_index].set(1.0),
                            rotation=state.joint.rotation.at[joint_index].set(b.rotation - a.rotation),
                        )
                    )

                    return state

                env_state = _add_joint_to_state(env_state)

                env_state = env_state.replace(
                    collision_matrix=calculate_collision_matrix(self.static_env_params, env_state.joint)
                )

        return env_state

    def _reset_select_shape(self, env_state):
        env_state = env_state.replace(
            polygon_highlighted=jnp.zeros_like(env_state.polygon_highlighted),
            circle_highlighted=jnp.zeros_like(env_state.circle_highlighted),
        )

        self.selected_shape_index = -1
        self.selected_shape_type = ObjectType.POLYGON
        self._hide_all_widgets()
        return env_state

    def _hide_all_widgets(self):
        for widget in self.all_widgets.values():
            for w in widget.values():
                w.hide()

    def _show_correct_widgets(self, type: ObjectType | None):
        for widget in self.all_widgets["GENERAL"].values():
            widget.show()

        for widget in self.all_widgets[type].values():
            widget.show()
        if type is None:
            self.all_widgets["GENERAL"]["lblGeneral"].setText(f"Global")
        else:
            self.all_widgets["GENERAL"]["lblGeneral"].setText(f"{type.name} (idx {self.selected_shape_index})")

    def _select_shape_keyboard_shortcuts(self, env_state: EnvState, left_click: bool, keys: list[int]):
        if left_click:
            return env_state
        if len(keys) != 0 and self.selected_shape_index != -1:
            s = 1.0
            ang_s = 0.1
            vel = jnp.array([0.0, 0.0])
            angular_vel = 0.0
            should_toggle_fixed = False
            should_toggle_collidable = False
            change_angle = 0

            def add_step(widget_name, direction, speed=10, overwrite_amount=None):
                widget = self.all_widgets[self.selected_shape_type][widget_name]
                val = widget.getValue()
                step = widget.step
                amount_to_add = overwrite_amount or step * direction * speed
                widget.setValue(jnp.clip(val + amount_to_add, widget.min, widget.max))

            if pygame.K_w in keys:
                add_step("sldPosition_Y", 1)
            if pygame.K_s in keys:
                add_step("sldPosition_Y", -1)
            if pygame.K_a in keys:
                add_step("sldPosition_X", -1)
            if pygame.K_d in keys:
                add_step("sldPosition_X", 1)

            if pygame.K_q in keys:
                add_step("sldRotation", 1)
            if pygame.K_e in keys:
                add_step("sldRotation", -1)

            if pygame.K_f in keys:
                self.all_widgets[self.selected_shape_type]["tglFixate"].toggle()

            if pygame.K_c in keys and not (pygame.key.get_mods() & pygame.KMOD_CTRL):
                widget = self.all_widgets[self.selected_shape_type]["sldCollidability"]
                curr_val = int(widget.getValue())
                widget.setValue((curr_val + 1) % (widget.max + 1))

            if pygame.K_r in keys and not (pygame.key.get_mods() & pygame.KMOD_CTRL):
                widget = self.all_widgets[self.selected_shape_type]["sldRole"]
                curr_val = int(widget.getValue())
                widget.setValue((curr_val + 1) % (widget.max + 1))

            if pygame.K_LEFTBRACKET in keys:
                add_step("sldRotation", 1, 10, jnp.pi / 4)

            if pygame.K_RIGHTBRACKET in keys:
                add_step("sldRotation", -1, 10, -jnp.pi / 4)

            if pygame.K_c in keys and (pygame.key.get_mods() & pygame.KMOD_CTRL):
                # copy
                if self.selected_shape_type == ObjectType.POLYGON:  # rect
                    if not self.env_state.polygon.active.all():
                        where_to_add = jnp.argmin(env_state.polygon.active)
                        if where_to_add < self.static_env_params.num_polygons:

                            def _copy_polygon(state, shift):
                                state = state.replace(
                                    polygon=jax.tree.map(
                                        lambda x: x.at[where_to_add].set(x[self.selected_shape_index]), state.polygon
                                    )
                                )
                                if shift:
                                    state = state.replace(
                                        polygon=state.polygon.replace(
                                            position=state.polygon.position.at[where_to_add].add(0.1),
                                        ),
                                        polygon_highlighted=state.polygon_highlighted.at[where_to_add].set(False),
                                    )
                                return state

                            env_state = _copy_polygon(env_state, shift=True)

                elif self.selected_shape_type == ObjectType.CIRCLE:  # circle
                    if not self.env_state.circle.active.all():
                        where_to_add = jnp.argmin(env_state.circle.active)
                        if where_to_add < self.static_env_params.num_circles:

                            def _copy_circle(state, shift=True):
                                state = state.replace(
                                    circle=jax.tree.map(
                                        lambda x: x.at[where_to_add].set(x[self.selected_shape_index]), state.circle
                                    )
                                )
                                if shift:
                                    state = state.replace(
                                        circle=state.circle.replace(
                                            position=state.circle.position.at[where_to_add].add(0.1),
                                        ),
                                        circle_highlighted=state.circle_highlighted.at[where_to_add].set(False),
                                    )
                                return state

                            env_state = _copy_circle(env_state)

            if self.selected_shape_index >= 0:
                num = get_numeric_key_pressed(self.pygame_events)
                if num is not None:
                    if self.selected_shape_type in [ObjectType.CIRCLE, ObjectType.POLYGON]:
                        self.all_widgets[self.selected_shape_type]["sldRole"].setValue(num % 4)
                    elif self.selected_shape_type == ObjectType.JOINT:
                        self.all_widgets[self.selected_shape_type]["sldColour"].setValue(
                            num % self.static_env_params.num_motor_bindings
                        )
                    elif self.selected_shape_type == ObjectType.THRUSTER:
                        self.all_widgets[self.selected_shape_type]["sldColour"].setValue(
                            num % self.static_env_params.num_thruster_bindings
                        )

        return env_state

    def _edit_select_shape(self, env_state: EnvState, left_click: bool, right_click: bool, keys: list[int]):
        def _find_shape(env_state):
            found_shape = False
            selected_shape_index, selected_shape_type = -1, ObjectType.POLYGON
            for ri in self._get_revolute_joints_on_mouse(env_state):
                selected_shape_index = ri
                selected_shape_type = ObjectType.JOINT
                found_shape = True
                break
            if not found_shape:
                for ti in self._get_thrusters_on_mouse(env_state):
                    selected_shape_index = ti
                    selected_shape_type = ObjectType.THRUSTER
                    found_shape = True
                    break
            if not found_shape:
                for ri in self._get_polygons_on_mouse(env_state):
                    env_state = env_state.replace(
                        polygon_highlighted=env_state.polygon_highlighted.at[ri].set(True),
                    )

                    selected_shape_index = ri
                    selected_shape_type = ObjectType.POLYGON
                    found_shape = True
                    break
            if not found_shape:
                for ci in self._get_circles_on_mouse(env_state):
                    env_state = env_state.replace(
                        circle_highlighted=env_state.circle_highlighted.at[ci].set(True),
                    )

                    selected_shape_index = ci
                    selected_shape_type = ObjectType.CIRCLE
                    found_shape = True
                    break
            return selected_shape_index, selected_shape_type, found_shape, env_state

        if left_click:
            self.all_selected_shapes = []
            self._hide_all_widgets()
            env_state = self._reset_select_shape(env_state)
            self.selected_shape_index, self.selected_shape_type, found_shape, env_state = _find_shape(env_state)
            if found_shape:
                self.all_selected_shapes = [(self.selected_shape_index, self.selected_shape_type)]
                if self.selected_shape_type in self.all_widgets:
                    self._show_correct_widgets(self.selected_shape_type)
        if self.selected_shape_index < 0:
            self._show_correct_widgets(None)

        return env_state

    def render(self, env_state):
        # Clear
        self.screen_surface.fill((0, 0, 0))

        if self.is_editing:
            pixels = self._render_fn_edit(env_state)
        else:
            pixels = self._render_fn(env_state)
        pixels = self._render_edit_overlay_fn(pixels, self.is_editing, self.edit_shape_mode.value)

        surface = pygame.surfarray.make_surface(np.array(pixels))
        self.screen_surface.blit(surface, (0, 0))

    def is_quit_requested(self):
        for event in self.pygame_events:
            if event.type == pygame.QUIT:
                return True
        return False


@hydra.main(version_base=None, config_path="../configs", config_name="editor")
def main(config):
    config = normalise_config(OmegaConf.to_container(config), "EDITOR", editor_config=True)
    env_params, static_env_params = generate_params_from_config(config)
    static_env_params = static_env_params.replace(frame_skip=config["frame_skip"], downscale=config["downscale"])
    config["env_params"] = to_state_dict(env_params)
    config["static_env_params"] = to_state_dict(static_env_params)

    env = make_kinetix_env(
        ActionType.MULTI_DISCRETE,
        ObservationType.SYMBOLIC_ENTITY,
        make_reset_function(static_env_params),
        env_params,
        static_env_params,
    )

    seed = config["seed"]
    print("seed", seed)
    rng = jax.random.PRNGKey(seed)
    editor = Editor(env, env_params, config, upscale=config["upscale"])

    clock = pygame.time.Clock()
    while not editor.is_quit_requested():
        rng, _rng = jax.random.split(rng)
        editor.update(_rng)
        clock.tick(config["fps"])


if __name__ == "__main__":
    main()



import jax
import jax.numpy as jnp
import numpy as np
from flax import struct
from jax2d.engine import select_shape
from jax2d.maths import rmat
from jaxgl.maths import dist_from_line
from jaxgl.renderer import clear_screen, make_renderer
from jaxgl.shaders import (
    nearest_neighbour,
    make_fragment_shader_convex_ngon_with_edges,
    fragment_shader_circle,
)

from kinetix.environment.env_state import EnvState, StaticEnvParams
from kinetix.render.textures import FJOINT_TEXTURE_6_RGBA, RJOINT_TEXTURE_6_RGBA, THRUSTER_TEXTURE_16_RGBA


def make_render_pixels(env_params, static_params: StaticEnvParams):
    screen_dim = static_params.screen_dim
    downscale = static_params.downscale

    joint_tex_size = 6
    thruster_tex_size = 16

    edge_thickness = 2
    use_textures_for_joints = True
    # if we downscale, the edges are too thick, and the textures do not show.
    if downscale == 2:
        edge_thickness = 1
    elif downscale == 4:
        edge_thickness = 0
        use_textures_for_joints = False
    fragment_shader_quad_no_edges = make_fragment_shader_convex_ngon_with_edges(4, edge_thickness=edge_thickness)

    FIXATED_COLOUR = jnp.array([80, 80, 80])
    JOINT_COLOURS = jnp.array(
        [
            # [0, 0, 255],
            [255, 255, 255],  # yellow
            [255, 255, 0],  # yellow
            [255, 0, 255],  # purple/magenta
            [0, 255, 255],  # cyan
            [255, 153, 51],  # white
        ]
    )

    def colour_thruster_texture(colour):
        return THRUSTER_TEXTURE_16_RGBA.at[:9, :, :3].mul(colour[None, None, :] / 255.0)

    coloured_thruster_textures = jax.vmap(colour_thruster_texture)(JOINT_COLOURS)

    ROLE_COLOURS = jnp.array(
        [
            [160.0, 160.0, 160.0],  # None
            [0.0, 204.0, 0.0],  # Green:    The ball
            [0.0, 102.0, 204.0],  # Blue:   The goal
            [255.0, 102.0, 102.0],  # Red:      Death Objects
        ]
    )

    BACKGROUND_COLOUR = jnp.array([255.0, 255.0, 255.0])

    def _get_colour(shape_role, inverse_inertia):
        base_colour = ROLE_COLOURS[shape_role]
        f = (inverse_inertia == 0) * 1
        is_not_normal = (shape_role != 0) * 1

        return jnp.array(
            [
                base_colour,
                base_colour,
                FIXATED_COLOUR,
                base_colour * 0.5,
            ]
        )[2 * f + is_not_normal]

    # Pixels per unit distance
    ppud = env_params.pixels_per_unit // downscale

    downscaled_screen_dim = (screen_dim[0] // downscale, screen_dim[1] // downscale)

    full_screen_size = (
        downscaled_screen_dim[0] + (static_params.max_shape_size * 2 * ppud),
        downscaled_screen_dim[1] + (static_params.max_shape_size * 2 * ppud),
    )
    cleared_screen = clear_screen(full_screen_size, BACKGROUND_COLOUR)

    def _world_space_to_pixel_space(x):
        return (x + static_params.max_shape_size) * ppud

    def fragment_shader_kinetix_circle(position, current_frag, unit_position, uniform):
        centre, radius, rotation, colour, mask = uniform

        dist = jnp.sqrt(jnp.square(position - centre).sum())
        inside = dist <= radius
        on_edge = dist > radius - edge_thickness

        normal = jnp.array([jnp.sin(rotation), -jnp.cos(rotation)])

        dist = dist_from_line(position, centre, centre + normal)

        on_edge |= (dist < 1) & (jnp.dot(normal, position - centre) <= 0)

        fragment = jax.lax.select(on_edge, jnp.zeros(3), colour)

        return jax.lax.select(inside & mask, fragment, current_frag)

    def fragment_shader_kinetix_joint_circle(position, current_frag, unit_position, uniform):
        centre, radius, colour, mask = uniform
        new = fragment_shader_circle(position, current_frag, unit_position, (centre, radius, colour))
        return new * mask + (1 - mask) * current_frag

    def fragment_shader_kinetix_joint_texture(position, current_frag, unit_position, uniform):
        texture, colour, mask = uniform

        tex_coord = (
            jnp.array(
                [
                    joint_tex_size * unit_position[0],
                    joint_tex_size * unit_position[1],
                ]
            )
            - 0.5
        )

        tex_frag = nearest_neighbour(texture, tex_coord)
        tex_frag = tex_frag.at[3].mul(mask)
        tex_frag = tex_frag.at[:3].mul(colour / 255.0)

        tex_frag = (tex_frag[3] * tex_frag[:3]) + ((1.0 - tex_frag[3]) * current_frag)

        return tex_frag

    thruster_pixel_size = thruster_tex_size // downscale
    thruster_pixel_size_diagonal = (thruster_pixel_size * np.sqrt(2)).astype(jnp.int32) + 1

    def fragment_shader_kinetix_thruster(fragment_position, current_frag, unit_position, uniform):
        thruster_position, rotation, texture, mask = uniform

        tex_position = jnp.matmul(rmat(-rotation), (fragment_position - thruster_position)) / thruster_pixel_size + 0.5

        mask &= (tex_position[0] >= 0) & (tex_position[0] <= 1) & (tex_position[1] >= 0) & (tex_position[1] <= 1)

        eps = 0.001
        tex_coord = (
            jnp.array(
                [
                    thruster_tex_size * tex_position[0],
                    thruster_tex_size * tex_position[1],
                ]
            )
            - 0.5
            + eps
        )

        tex_frag = nearest_neighbour(texture, tex_coord)
        tex_frag = tex_frag.at[3].mul(mask)

        tex_frag = (tex_frag[3] * tex_frag[:3]) + ((1.0 - tex_frag[3]) * current_frag)

        return tex_frag

    patch_size_1d = static_params.max_shape_size * ppud
    patch_size = (patch_size_1d, patch_size_1d)

    circle_renderer = make_renderer(full_screen_size, fragment_shader_kinetix_circle, patch_size, batched=True)
    quad_renderer = make_renderer(full_screen_size, fragment_shader_quad_no_edges, patch_size, batched=True)
    big_quad_renderer = make_renderer(full_screen_size, fragment_shader_quad_no_edges, downscaled_screen_dim)

    joint_pixel_size = joint_tex_size // downscale
    joint_fragment_shader = (
        fragment_shader_kinetix_joint_texture if use_textures_for_joints else fragment_shader_kinetix_joint_circle
    )
    joint_renderer = make_renderer(
        full_screen_size, joint_fragment_shader, (joint_pixel_size, joint_pixel_size), batched=True
    )

    thruster_renderer = make_renderer(
        full_screen_size,
        fragment_shader_kinetix_thruster,
        (thruster_pixel_size_diagonal, thruster_pixel_size_diagonal),
        batched=True,
    )

    @jax.jit
    def render_pixels(state: EnvState):
        pixels = cleared_screen

        # Floor
        floor_uniform = (
            _world_space_to_pixel_space(state.polygon.position[0, None, :] + state.polygon.vertices[0]),
            _get_colour(state.polygon_shape_roles[0], 0),
            jnp.zeros(3),
            True,
        )

        pixels = big_quad_renderer(pixels, _world_space_to_pixel_space(jnp.zeros(2, dtype=jnp.int32)), floor_uniform)

        # Rectangles
        rectangle_patch_positions = _world_space_to_pixel_space(
            state.polygon.position - (static_params.max_shape_size / 2.0)
        ).astype(jnp.int32)

        rectangle_rmats = jax.vmap(rmat)(state.polygon.rotation)
        rectangle_rmats = jnp.repeat(rectangle_rmats[:, None, :, :], repeats=static_params.max_polygon_vertices, axis=1)
        rectangle_vertices_pixel_space = _world_space_to_pixel_space(
            state.polygon.position[:, None, :] + jax.vmap(jax.vmap(jnp.matmul))(rectangle_rmats, state.polygon.vertices)
        )
        rectangle_colours = jax.vmap(_get_colour)(state.polygon_shape_roles, state.polygon.inverse_mass)
        rectangle_edge_colours = jnp.zeros((static_params.num_polygons, 3))

        rectangle_uniforms = (
            rectangle_vertices_pixel_space,
            rectangle_colours,
            rectangle_edge_colours,
            state.polygon.active.at[: static_params.num_static_fixated_polys].set(False),
        )

        pixels = quad_renderer(pixels, rectangle_patch_positions, rectangle_uniforms)

        # Circles
        circle_positions_pixel_space = _world_space_to_pixel_space(state.circle.position)
        circle_radii_pixel_space = state.circle.radius * ppud
        circle_patch_positions = _world_space_to_pixel_space(
            state.circle.position - (static_params.max_shape_size / 2.0)
        ).astype(jnp.int32)

        circle_colours = jax.vmap(_get_colour)(state.circle_shape_roles, state.circle.inverse_mass)

        circle_uniforms = (
            circle_positions_pixel_space,
            circle_radii_pixel_space,
            state.circle.rotation,
            circle_colours,
            state.circle.active,
        )

        pixels = circle_renderer(pixels, circle_patch_positions, circle_uniforms)

        # Joints
        joint_patch_positions = jnp.round(
            _world_space_to_pixel_space(state.joint.global_position) - (joint_pixel_size // 2)
        ).astype(jnp.int32)
        joint_textures = jax.vmap(jax.lax.select, in_axes=(0, None, None))(
            state.joint.is_fixed_joint, FJOINT_TEXTURE_6_RGBA, RJOINT_TEXTURE_6_RGBA
        )
        joint_colours = JOINT_COLOURS[
            (state.motor_bindings + 1) * (state.joint.motor_on & (~state.joint.is_fixed_joint))
        ]

        if use_textures_for_joints:
            joint_uniforms = (joint_textures, joint_colours, state.joint.active)
        else:
            joint_uniforms = (
                joint_patch_positions + joint_pixel_size // 2,
                jnp.ones(static_params.num_joints) * joint_pixel_size / 2,
                joint_colours.astype(jnp.float32),
                state.joint.active,
            )

        pixels = joint_renderer(pixels, joint_patch_positions, joint_uniforms)

        # Thrusters
        thruster_positions = jnp.round(_world_space_to_pixel_space(state.thruster.global_position)).astype(jnp.int32)
        thruster_patch_positions = thruster_positions - (thruster_pixel_size_diagonal // 2)
        thruster_textures = coloured_thruster_textures[state.thruster_bindings + 1]
        thruster_rotations = (
            state.thruster.rotation
            + jax.vmap(select_shape, in_axes=(None, 0, None))(
                state, state.thruster.object_index, static_params
            ).rotation
        )
        thruster_uniforms = (thruster_positions, thruster_rotations, thruster_textures, state.thruster.active)

        pixels = thruster_renderer(pixels, thruster_patch_positions, thruster_uniforms)

        # Crop out the sides
        crop_amount = static_params.max_shape_size * ppud
        return pixels[crop_amount:-crop_amount, crop_amount:-crop_amount]

    return render_pixels


@struct.dataclass
class PixelsObservation:
    image: jnp.ndarray
    global_info: jnp.ndarray


def make_render_pixels_rl(env_params, static_params: StaticEnvParams):
    render_fn = make_render_pixels(env_params, static_params)

    def inner(state):
        pixels = render_fn(state) / 255.0
        return PixelsObservation(
            image=pixels,
            global_info=jnp.array([state.gravity[1] / 10.0]),
        )

    return inner
